//
// $Source: /cvsroot/gambit/gambit/sources/game/Attic/efg.cc,v $
// $Date: 2005/06/13 02:32:09 $
// $Revision: 1.7.2.17 $
//
// DESCRIPTION:
// Implementation of extensive form game representation
//
// This file is part of Gambit
// Copyright (c) 2002, The Gambit Project
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
//

#include <iostream>

#include "base/base.h"
#include "math/rational.h"

#if defined(__GNUG__) && !defined(__APPLE_CC__)
#pragma implementation "infoset.h"
#pragma implementation "efplayer.h"
#pragma implementation "node.h"
#endif   // __GNUG__

#include "efg.h"
#include "efgutils.h"
#include "efstrat.h"
#include "nfg.h"

static std::string EscapeQuotes(const std::string &s)
{
  std::string ret;
  
  for (unsigned int i = 0; i < s.length(); i++)  {
    if (s[i] == '"')   ret += '\\';
    ret += s[i];
  }

  return ret;
}


//----------------------------------------------------------------------
//                 gbtEfgPlayer: Member function definitions
//----------------------------------------------------------------------

gbtEfgPlayer::~gbtEfgPlayer()
{
  while (infosets.Length())   delete infosets.Remove(1);
}


//----------------------------------------------------------------------
//                 gbtEfgAction: Member function definitions
//----------------------------------------------------------------------

bool gbtEfgAction::Precedes(const gbtEfgNode * n) const
{
  while ( n != n->Game()->RootNode() ) {
    if ( n->GetAction() == this )
      return true;
    else
      n = n->GetParent();
  }
  return false;
}

//----------------------------------------------------------------------
//                 Infoset: Member function definitions
//----------------------------------------------------------------------

bool gbtEfgInfoset::Precedes(const gbtEfgNode * n) const
{
  while ( n != n->Game()->RootNode() ) {
    if ( n->GetInfoset() == this )
      return true;
    else
      n = n->GetParent();
  }
  return false;
}

gbtEfgInfoset::gbtEfgInfoset(gbtEfgGame *e, int n, gbtEfgPlayer *p, int br)
  : E(e), number(n), player(p), actions(br), flag(0) 
{
  while (br)   {
    actions[br] = new gbtEfgAction(br, "", this);
    br--; 
  }
}

gbtEfgInfoset::~gbtEfgInfoset()  
{
  for (int i = 1; i <= actions.Length(); i++)  delete actions[i];
}

void gbtEfgInfoset::PrintActions(std::ostream &f) const
{ 
  f << "{ ";
  for (int i = 1; i <= actions.Length(); i++)
    f << '"' << EscapeQuotes(actions[i]->name) << "\" ";
  f << "}";
}

gbtEfgAction *gbtEfgInfoset::InsertAction(int where)
{
  gbtEfgAction *action = new gbtEfgAction(where, "", this);
  actions.Insert(action, where);
  for (; where <= actions.Length(); where++)
    actions[where]->number = where;
  return action;
}

void gbtEfgInfoset::RemoveAction(int which)
{
  delete actions.Remove(which);
  for (; which <= actions.Length(); which++)
    actions[which]->number = which;
}

const gList<gbtEfgAction *> gbtEfgInfoset::ListOfActions(void) const
{
  gList<gbtEfgAction *> answer;
  for (int i = 1; i <= actions.Length(); i++) 
    answer.Append(actions[i]);
  return answer;
}

const gList<const gbtEfgNode *> gbtEfgInfoset::ListOfMembers(void) const
{
  gList<const gbtEfgNode *> answer;
  for (int i = 1; i <= members.Length(); i++) 
    answer.Append(members[i]);
  return answer;
}

//------------------------------------------------------------------------
//           ChanceInfoset: Member function definitions
//------------------------------------------------------------------------

ChanceInfoset::ChanceInfoset(gbtEfgGame *E, int n, gbtEfgPlayer *p, int br)
  : gbtEfgInfoset(E, n, p, br), m_textProbs(br), m_ratProbs(br)
{
  for (int act = 1; act <= br; act++) {
    m_ratProbs[act] = gRational(1, br);
    m_textProbs[act] = ToText(m_ratProbs[act]);
  }
}

gbtEfgAction *ChanceInfoset::InsertAction(int where)
{ 
  gbtEfgAction *action = gbtEfgInfoset::InsertAction(where);
  m_textProbs.Insert("0", where);
  m_ratProbs.Insert(gRational(0), where);
  return action;
}

void ChanceInfoset::RemoveAction(int which)
{
  gbtEfgInfoset::RemoveAction(which);
  m_textProbs.Remove(which);
  m_ratProbs.Remove(which);
}

void ChanceInfoset::PrintActions(std::ostream &f) const
{ 
  f << "{ ";
  for (int i = 1; i <= actions.Length(); i++) {
    f << '"' << EscapeQuotes(actions[i]->GetName()) << "\" ";
    f << m_textProbs[i] << ' ';
  }
  f << "}";
}

//----------------------------------------------------------------------
//                 gbtEfgNode: Member function definitions
//----------------------------------------------------------------------

gbtEfgNode::gbtEfgNode(gbtEfgGame *e, gbtEfgNode *p)
  : mark(false), number(0), E(e), infoset(0), parent(p), outcome(0),
    gameroot((p) ? p->gameroot : this)
{ }

gbtEfgNode::~gbtEfgNode()
{
  for (int i = children.Length(); i; delete children[i--]);
}


int gbtEfgNode::NumberInInfoset(void) const
{
  for (int i = 1; i <= GetInfoset()->NumMembers(); i++)
    if (GetInfoset()->GetMember(i) == this)
      return i;
  //  This could be speeded up by adding a member to Node to keep track of this
  throw gbtEfgGame::Exception();
}


gbtEfgNode *gbtEfgNode::NextSibling(void) const  
{
  if (!parent)   return 0;
  if (parent->children.Find((gbtEfgNode * const) this) == parent->children.Length())
    return 0;
  else
    return parent->children[parent->children.Find((gbtEfgNode * const)this) + 1];
}

gbtEfgNode *gbtEfgNode::PriorSibling(void) const
{ 
  if (!parent)   return 0;
  if (parent->children.Find((gbtEfgNode * const)this) == 1)
    return 0;
  else
    return parent->children[parent->children.Find((gbtEfgNode * const)this) - 1];

}

gbtEfgAction *gbtEfgNode::GetAction(void) const
{
  if (this == Game()->RootNode()) {
    return 0;
  }
  
  const gArray<gbtEfgAction *> &actions = GetParent()->GetInfoset()->Actions();
  for (int i = 1; i <= actions.Length(); i++) {
    if (this == GetParent()->GetChild(actions[i])) {
      return actions[i];
    }
  }

  return 0;
}

void gbtEfgNode::DeleteOutcome(gbtEfgOutcome *outc)
{
  if (outc == outcome)   outcome = 0;
  for (int i = 1; i <= children.Length(); i++)
    children[i]->DeleteOutcome(outc);
}

void gbtEfgNode::SetOutcome(gbtEfgOutcome *p_outcome)
{
  if (outcome != p_outcome) {
    outcome = p_outcome;
    E->DeleteLexicon();
  }
}

//------------------------------------------------------------------------
//       Efg: Constructors, destructor, constructive operators
//------------------------------------------------------------------------

gbtEfgGame::gbtEfgGame(void)
  : sortisets(true), title("Untitled extensive form game"),
    chance(new gbtEfgPlayer(this, 0)), lexicon(0)
{
  root = new gbtEfgNode(this, 0);
  SortInfosets();
}

gbtEfgGame::gbtEfgGame(const gbtEfgGame &E, gbtEfgNode *n /* = 0 */)
  : sortisets(false), title(E.title), comment(E.comment),
    players(E.players.Length()), outcomes(0, E.outcomes.Last()),
    chance(new gbtEfgPlayer(this, 0)), lexicon(0) 
{
  for (int i = 1; i <= players.Length(); i++)  {
    (players[i] = new gbtEfgPlayer(this, i))->name = E.players[i]->name;
    for (int j = 1; j <= E.players[i]->infosets.Length(); j++)   {
      gbtEfgInfoset *s = new gbtEfgInfoset(this, j, players[i],
			       E.players[i]->infosets[j]->actions.Length());
      s->name = E.players[i]->infosets[j]->name;
      for (int k = 1; k <= s->actions.Length(); k++)
	s->actions[k]->name = E.players[i]->infosets[j]->actions[k]->name;
      players[i]->infosets.Append(s);
    }
  }

  for (int i = 1; i <= E.GetChance()->NumInfosets(); i++)   {
    ChanceInfoset *t = (ChanceInfoset *) E.GetChance()->Infosets()[i];
    ChanceInfoset *s = new ChanceInfoset(this, i, chance,
					       t->NumActions());
    s->name = t->GetName();
    for (int act = 1; act <= s->m_textProbs.Length(); act++) {
      s->m_textProbs[act] = t->m_textProbs[act];
      s->m_ratProbs[act] = t->m_ratProbs[act];
      s->actions[act]->name = t->actions[act]->name;
    }
    chance->infosets.Append(s);
  }

  for (int outc = 1; outc <= E.NumOutcomes(); outc++)  {
    outcomes[outc] = new gbtEfgOutcome(this, outc);
    outcomes[outc]->m_label = E.outcomes[outc]->m_label;
    outcomes[outc]->m_textPayoffs = E.outcomes[outc]->m_textPayoffs;
    outcomes[outc]->m_ratPayoffs = E.outcomes[outc]->m_ratPayoffs;
    outcomes[outc]->m_doublePayoffs = E.outcomes[outc]->m_doublePayoffs;
  }

  root = new gbtEfgNode(this, 0);
  CopySubtree(root, (n ? n : E.RootNode()));
  
  if (n)   {
    for (int pl = 1; pl <= players.Length(); pl++)  {
      for (int i = 1; i <= players[pl]->infosets.Length(); i++)  {
	if (players[pl]->infosets[i]->members.Length() == 0)
	  delete players[pl]->infosets.Remove(i--);
      }
    }
  }

  sortisets = true;
  SortInfosets();
}

#include "lexicon.h"

gbtEfgGame::~gbtEfgGame()
{
  delete root;
  delete chance;

  for (int i = 1; i <= players.Length(); delete players[i++]);
  for (int i = 1; i <= outcomes.Last(); delete outcomes[i++]);

  if (lexicon) {
    // Note this behavior change from previous versions:
    // We will delete the normal form!
    // This is an intermediate stage; we will eventually
    // be backporting the "implicit" nfg representation code from
    // the development version (in which case we won't even be
    // physically creating a normal form!)

    if (lexicon->N)  {
      lexicon->N->efg = 0;
      //delete lexicon->N;
    }
    delete lexicon;
  }
  lexicon = 0;
}

//------------------------------------------------------------------------
//                  Efg: Private member functions
//------------------------------------------------------------------------

void gbtEfgGame::DeleteLexicon(void) const
{
  if (!lexicon)  return;

  // Note this behavior change from previous versions:
  // We will delete the normal form!
  // This is an intermediate stage; we will eventually
  // be backporting the "implicit" nfg representation code from
  // the development version (in which case we won't even be
  // physically creating a normal form!)
  if (lexicon->N)  {
    lexicon->N->efg = 0;
    //delete lexicon->N;
  }
  delete lexicon;
  lexicon = 0;
}

gbtEfgInfoset *gbtEfgGame::GetInfosetByIndex(gbtEfgPlayer *p, int index) const
{
  for (int i = 1; i <= p->infosets.Length(); i++)
    if (p->infosets[i]->number == index)   return p->infosets[i];
  return 0;
}

gbtEfgOutcome *gbtEfgGame::GetOutcomeByIndex(int index) const
{
  for (int i = 1; i <= outcomes.Last(); i++) {
    if (outcomes[i]->m_number == index)  {
      return outcomes[i];
    }
  }

  return 0;
}

void gbtEfgGame::Reindex(void)
{
  int i;

  for (i = 1; i <= players.Length(); i++)  {
    gbtEfgPlayer *p = players[i];
    for (int j = 1; j <= p->infosets.Length(); j++)
      p->infosets[j]->number = j;
  }

  for (i = 1; i <= outcomes.Last(); i++)
    outcomes[i]->m_number = i;
}


void gbtEfgGame::NumberNodes(gbtEfgNode *n, int &index)
{
  n->number = index++;
  for (int child = 1; child <= n->children.Length();
       NumberNodes(n->children[child++], index));
} 

void gbtEfgGame::SortInfosets(void)
{
  if (!sortisets)  return;

  int pl;

  for (pl = 0; pl <= players.Length(); pl++)  {
    gList<gbtEfgNode *> nodes;

    Nodes(*this, nodes);

    gbtEfgPlayer *player = (pl) ? players[pl] : chance;

    int i, isets = 0;

    // First, move all empty infosets to the back of the list so
    // we don't "lose" them
    int foo = player->infosets.Length();
    i = 1;
    while (i < foo)   {
      if (player->infosets[i]->members.Length() == 0)  {
	gbtEfgInfoset *bar = player->infosets[i];
	player->infosets[i] = player->infosets[foo];
	player->infosets[foo--] = bar;
      }
      else
	i++;
    }

    // This will give empty infosets their proper number; the nonempty
    // ones will be renumbered by the next loop
    for (i = 1; i <= player->infosets.Length(); i++)
      if (player->infosets[i]->members.Length() == 0)
	player->infosets[i]->number = i;
      else
	player->infosets[i]->number = 0;
  
    for (i = 1; i <= nodes.Length(); i++)  {
      gbtEfgNode *n = nodes[i];
      if (n->GetPlayer() == player && n->GetInfoset()->number == 0)  {
	n->GetInfoset()->number = ++isets;
	player->infosets[isets] = n->GetInfoset();
      }
    }  
  }

  // Now, we sort the nodes within the infosets
  
  gList<gbtEfgNode *> nodes;
  Nodes(*this, nodes);

  for (pl = 0; pl <= players.Length(); pl++)  {
    gbtEfgPlayer *player = (pl) ? players[pl] : chance;

    for (int iset = 1; iset <= player->infosets.Length(); iset++)  {
      gbtEfgInfoset *s = player->infosets[iset];
      for (int i = 1, j = 1; i <= nodes.Length(); i++)  {
	if (nodes[i]->infoset == s)
	  s->members[j++] = nodes[i];
      }
    }
  }

  int nodeindex = 1;
  NumberNodes(root, nodeindex);
}
  
gbtEfgInfoset *gbtEfgGame::CreateInfoset(int n, gbtEfgPlayer *p, int br)
{
  gbtEfgInfoset *s = (p->IsChance()) ? new ChanceInfoset(this, n, p, br) :
               new gbtEfgInfoset(this, n, p, br);
  p->infosets.Append(s);
  return s;
}

gbtEfgOutcome *gbtEfgGame::CreateOutcomeByIndex(int index)
{
  NewOutcome(index);
  return outcomes[outcomes.Last()];
}

void gbtEfgGame::CopySubtree(gbtEfgNode *n, gbtEfgNode *m)
{
  n->m_label = m->m_label;

  if (m->gameroot == m)
    n->gameroot = n;

  if (m->outcome) {
    n->outcome = m->outcome;
  }

  if (m->infoset)   {
    gbtEfgPlayer *p;
    if (m->infoset->player->number)
      p = players[m->infoset->player->number];
    else
      p = chance;

    gbtEfgInfoset *s = p->Infosets()[m->infoset->number];
    AppendNode(n, s);

    for (int i = 1; i <= n->children.Length(); i++)
      CopySubtree(n->children[i], m->children[i]);
  }
}

//------------------------------------------------------------------------
//               Efg: Title access and manipulation
//------------------------------------------------------------------------

void gbtEfgGame::SetTitle(const std::string &s)
{
  title = s; 
}

const std::string &gbtEfgGame::GetTitle(void) const
{ return title; }

void gbtEfgGame::SetComment(const std::string &s)
{
  comment = s;
}

const std::string &gbtEfgGame::GetComment(void) const
{ return comment; }
  

//------------------------------------------------------------------------
//                    Efg: Writing data files
//------------------------------------------------------------------------

void gbtEfgGame::WriteEfgFile(std::ostream &f, gbtEfgNode *n) const
{
  if (n->children.Length() == 0)   {
    f << "t \"" << EscapeQuotes(n->m_label) << "\" ";
    if (n->outcome)  {
      f << n->outcome->m_number << " \"" <<
	EscapeQuotes(n->outcome->m_label) << "\" ";
      f << "{ ";
      for (int pl = 1; pl <= NumPlayers(); pl++)  {
	f << n->outcome->m_textPayoffs[pl];

	if (pl < NumPlayers())
	  f << ", ";
	else
	  f << " }\n";
      }
    }
    else
      f << "0\n";
  }

  else if (n->infoset->player->number)   {
    f << "p \"" << EscapeQuotes(n->m_label) << "\" " <<
      n->infoset->player->number << ' ';
    f << n->infoset->number << " \"" <<
      EscapeQuotes(n->infoset->name) << "\" ";
    n->infoset->PrintActions(f);
    f << " ";
    if (n->outcome)  {
      f << n->outcome->m_number << " \"" <<
	EscapeQuotes(n->outcome->m_label) << "\" ";
      f << "{ ";
      for (int pl = 1; pl <= NumPlayers(); pl++)  {
	f << n->outcome->m_textPayoffs[pl];

	if (pl < NumPlayers())
	  f << ", ";
	else
	  f << " }\n";
      }
    }
    else
      f << "0\n";
  }

  else   {    // chance node
    f << "c \"" << n->m_label << "\" ";
    f << n->infoset->number << " \"" <<
      EscapeQuotes(n->infoset->name) << "\" ";
    n->infoset->PrintActions(f);
    f << " ";
    if (n->outcome)  {
      f << n->outcome->m_number << " \"" <<
	EscapeQuotes(n->outcome->m_label) << "\" ";
      f << "{ ";
      for (int pl = 1; pl <= NumPlayers(); pl++)  {
	f << n->outcome->m_textPayoffs[pl];

        if (pl < NumPlayers()) 
          f << ", ";
        else
          f << " }\n";
      }
    }
    else
      f << "0\n";
  }

  for (int i = 1; i <= n->children.Length(); i++)
    WriteEfgFile(f, n->children[i]);
}

void gbtEfgGame::WriteEfgFile(std::ostream &p_file) const
{
  p_file << "EFG 2 R";
  p_file << " \"" << EscapeQuotes(title) << "\" { ";
  for (int i = 1; i <= players.Length(); i++)
    p_file << '"' << EscapeQuotes(players[i]->name) << "\" ";
  p_file << "}\n";
  p_file << "\"" << EscapeQuotes(comment) << "\"\n\n";

  WriteEfgFile(p_file, root);
}


//------------------------------------------------------------------------
//                    Efg: General data access
//------------------------------------------------------------------------

int gbtEfgGame::NumPlayers(void) const
{ return players.Length(); }

gbtEfgPlayer *gbtEfgGame::NewPlayer(void)
{
  gbtEfgPlayer *ret = new gbtEfgPlayer(this, players.Length() + 1);
  players.Append(ret);

  for (int outc = 1; outc <= outcomes.Last(); outc++) {
    outcomes[outc]->m_textPayoffs.Append("0");
    outcomes[outc]->m_ratPayoffs.Append(0);
    outcomes[outc]->m_doublePayoffs.Append(0.0);
  }

  DeleteLexicon();
  return ret;
}

gArray<gbtEfgInfoset *> gbtEfgGame::Infosets() const
{
  gArray<gbtEfgInfoset *> answer;

  gArray<gbtEfgPlayer *> p = Players();
  int i;
  for (i = 1; i <= p.Length(); i++) {
    gArray<gbtEfgInfoset *> infosets_for_player = p[i]->Infosets();
    int j;
    for (j = 1; j <= infosets_for_player.Length(); j++)
      answer.Append(infosets_for_player[j]);
  }

  return answer;
}

int gbtEfgGame::NumOutcomes(void) const
{ return outcomes.Last(); }

gbtEfgOutcome *gbtEfgGame::NewOutcome(void)
{
  return NewOutcome(outcomes.Last() + 1);
}

void gbtEfgGame::DeleteOutcome(gbtEfgOutcome *p_outcome)
{
  root->DeleteOutcome(p_outcome);
  delete outcomes.Remove(outcomes.Find(p_outcome));
  DeleteLexicon();
}

gbtEfgOutcome *gbtEfgGame::GetOutcome(int p_index) const
{
  return outcomes[p_index];
}

void gbtEfgGame::SetPayoff(gbtEfgOutcome *p_outcome, int pl, 
			   const std::string &p_value)
{
  if (!p_outcome) return;

  p_outcome->m_textPayoffs[pl] = p_value;
  // Note that ToRational() converts a decimal text string into
  // an exact fraction
  p_outcome->m_ratPayoffs[pl] = ToRational(p_value);
  p_outcome->m_doublePayoffs[pl] = (double) p_outcome->m_ratPayoffs[pl];
  DeleteLexicon();
}

gRational gbtEfgGame::Payoff(gbtEfgOutcome *p_outcome,
			     const gbtEfgPlayer *p_player) const
{
  if (!p_outcome) {
    return gRational(0);
  }

  return p_outcome->m_ratPayoffs[p_player->number];
}

std::string gbtEfgGame::GetPayoffText(gbtEfgOutcome *p_outcome,
				      const gbtEfgPlayer *p_player) const
{
  if (!p_outcome)  return "";
  return p_outcome->m_textPayoffs[p_player->number];
}

gRational gbtEfgGame::Payoff(const gbtEfgNode *p_node,
			     const gbtEfgPlayer *p_player) const
{
  if (p_node->outcome) {
    return p_node->outcome->m_ratPayoffs[p_player->number];
  }
  else {
    return gRational(0);
  }
}

gArray<gRational> gbtEfgGame::Payoff(gbtEfgOutcome *p_outcome) const
{
  if (!p_outcome) {
    gArray<gRational> ret(players.Length());
    for (int i = 1; i <= ret.Length(); ret[i++] = 0);
    return ret;
  }
  else {
    return p_outcome->m_ratPayoffs;
  }
}

bool gbtEfgGame::IsConstSum(void) const
{
  int pl, index;
  gRational cvalue(0);

  if (outcomes.Last() == 0)  return true;

  for (pl = 1; pl <= players.Length(); pl++)
    cvalue += outcomes[1]->m_ratPayoffs[pl];

  for (index = 2; index <= outcomes.Last(); index++)  {
    gRational thisvalue(0);

    for (pl = 1; pl <= players.Length(); pl++)
      thisvalue += outcomes[index]->m_ratPayoffs[pl];

    if (thisvalue > cvalue || thisvalue < cvalue)
      return false;
  }

  return true;
}

gRational gbtEfgGame::MinPayoff(int pl) const
{
  int index, p, p1, p2;
  gRational minpay;

  if (NumOutcomes() == 0)  return 0;

  if(pl) { p1=p2=pl;}
  else {p1=1;p2=players.Length();}

  minpay = outcomes[1]->m_ratPayoffs[p1];

  for (index = 1; index <= outcomes.Last(); index++)  {
    for (p = p1; p <= p2; p++)
      if (outcomes[index]->m_ratPayoffs[p] < minpay)
	minpay = outcomes[index]->m_ratPayoffs[p];
  }
  return minpay;
}

gRational gbtEfgGame::MaxPayoff(int pl) const
{
  int index, p, p1, p2;
  gRational maxpay;

  if (NumOutcomes() == 0)  return 0;

  if(pl) { p1=p2=pl;}
  else {p1=1;p2=players.Length();}

  maxpay = outcomes[1]->m_ratPayoffs[p1];

  for (index = 1; index <= outcomes.Last(); index++)  {
    for (p = p1; p <= p2; p++)
      if (outcomes[index]->m_ratPayoffs[p] > maxpay)
	maxpay = outcomes[index]->m_ratPayoffs[p];
  }
  return maxpay;
}

gbtEfgNode *gbtEfgGame::RootNode(void) const
{ return root; }

bool gbtEfgGame::IsSuccessor(const gbtEfgNode *n, const gbtEfgNode *from) const
{ return IsPredecessor(from, n); }

bool gbtEfgGame::IsPredecessor(const gbtEfgNode *n, const gbtEfgNode *of) const
{
  while (of && n != of)    of = of->parent;

  return (n == of);
}

gArray<int> gbtEfgGame::PathToNode(const gbtEfgNode *p_node) const
{
  gArray<int> ret;
  const gbtEfgNode *n = p_node;
  while (n->GetParent()) {
    ret.Insert(p_node->GetAction()->GetNumber(), 1);
    n = n->GetParent();
  }
  return ret;
}

void gbtEfgGame::DescendantNodes(const gbtEfgNode* n, 
			      const gbtEfgSupport& supp,
			      gList<gbtEfgNode *> &current) const
{
  current.Append(const_cast<gbtEfgNode *>(n));
  if (n->IsNonterminal()) {
    const gArray<gbtEfgAction *> actions = supp.Actions(n->GetInfoset());
    for (int i = 1; i <= actions.Length(); i++) {
      const gbtEfgNode* newn = n->GetChild(actions[i]);
      DescendantNodes(newn,supp,current);
    }
  }
}

void gbtEfgGame::NonterminalDescendants(const gbtEfgNode* n, 
					  const gbtEfgSupport& supp,
					  gList<const gbtEfgNode*>& current) const
{
  if (n->IsNonterminal()) {
    current.Append(n);
    const gArray<gbtEfgAction *> actions = supp.Actions(n->GetInfoset());
    for (int i = 1; i <= actions.Length(); i++) {
      const gbtEfgNode* newn = n->GetChild(actions[i]);
      NonterminalDescendants(newn,supp,current);
    }
  }
}

void gbtEfgGame::TerminalDescendants(const gbtEfgNode* n, 
				  const gbtEfgSupport& supp,
				  gList<gbtEfgNode *> &current) const
{
  if (n->IsTerminal()) { 
    current.Append(const_cast<gbtEfgNode *>(n));
  }
  else {
    const gArray<gbtEfgAction *> actions = supp.Actions(n->GetInfoset());
    for (int i = 1; i <= actions.Length(); i++) {
      const gbtEfgNode* newn = n->GetChild(actions[i]);
      TerminalDescendants(newn,supp,current);
    }
  }
}

gList<gbtEfgNode *>
gbtEfgGame::DescendantNodes(const gbtEfgNode &p_node, const gbtEfgSupport &p_support) const
{
  gList<gbtEfgNode *> answer;
  DescendantNodes(&p_node, p_support, answer);
  return answer;
}

gList<const gbtEfgNode*> 
gbtEfgGame::NonterminalDescendants(const gbtEfgNode& n, const gbtEfgSupport& supp) const
{
  gList<const gbtEfgNode*> answer;
  NonterminalDescendants(&n,supp,answer);
  return answer;
}

gList<gbtEfgNode *> 
gbtEfgGame::TerminalDescendants(const gbtEfgNode& n, const gbtEfgSupport& supp) const
{
  gList<gbtEfgNode *> answer;
  TerminalDescendants(&n,supp,answer);
  return answer;
}

gList<gbtEfgNode *> gbtEfgGame::TerminalNodes() const
{
  return TerminalDescendants(*(RootNode()),gbtEfgSupport(*this));
}

gList<gbtEfgInfoset*> gbtEfgGame::DescendantInfosets(const gbtEfgNode& n, 
					const gbtEfgSupport& supp) const
{
  gList<gbtEfgInfoset*> answer;
  gList<const gbtEfgNode*> nodelist = NonterminalDescendants(n,supp);
  int i;
  for (i = 1; i <= nodelist.Length(); i++) {
      gbtEfgInfoset* iset = nodelist[i]->GetInfoset();
      if (!answer.Contains(iset))
	answer.Append(iset);
  }
  return answer;
}

gbtEfgOutcome *gbtEfgGame::NewOutcome(int index)
{
  outcomes.Append(new gbtEfgOutcome(this, index));
  return outcomes[outcomes.Last()];
} 

//------------------------------------------------------------------------
//                     Efg: Operations on players
//------------------------------------------------------------------------

gbtEfgPlayer *gbtEfgGame::GetChance(void) const
{
  return chance;
}

gbtEfgInfoset *gbtEfgGame::AppendNode(gbtEfgNode *n, gbtEfgPlayer *p, int count)
{
  if (!n || !p || count == 0)
    throw Exception();

  if (n->children.Length() == 0)   {
    n->infoset = CreateInfoset(p->infosets.Length() + 1, p, count);
    n->infoset->members.Append(n);
    while (count--)
      n->children.Append(new gbtEfgNode(this, n));
  }

  DeleteLexicon();
  SortInfosets();
  return n->infoset;
}  

gbtEfgInfoset *gbtEfgGame::AppendNode(gbtEfgNode *n, gbtEfgInfoset *s)
{
  if (!n || !s)   throw Exception();
  
  // Can't bridge subgames...
  if (s->members.Length() > 0 && n->gameroot != s->members[1]->gameroot)
    return 0;

  if (n->children.Length() == 0)   {
    n->infoset = s;
    s->members.Append(n);
    for (int i = 1; i <= s->actions.Length(); i++)
      n->children.Append(new gbtEfgNode(this, n));
  }

  DeleteLexicon();
  SortInfosets();
  return s;
}
  
gbtEfgNode *gbtEfgGame::DeleteNode(gbtEfgNode *n, gbtEfgNode *keep)
{
  if (!n || !keep)   throw Exception();

  if (keep->parent != n)   return n;

  if (n->gameroot == n)
    MarkSubgame(keep, keep);

  // turn infoset sorting off during tree deletion -- problems will occur
  sortisets = false;

  n->children.Remove(n->children.Find(keep));
  DeleteTree(n);
  keep->parent = n->parent;
  if (n->parent)
    n->parent->children[n->parent->children.Find(n)] = keep;
  else
    root = keep;

  delete n;
  DeleteLexicon();

  sortisets = true;

  SortInfosets();
  return keep;
}

gbtEfgInfoset *gbtEfgGame::InsertNode(gbtEfgNode *n, gbtEfgPlayer *p, int count)
{
  if (!n || !p || count <= 0)  throw Exception();

  gbtEfgNode *m = new gbtEfgNode(this, n->parent);
  m->infoset = CreateInfoset(p->infosets.Length() + 1, p, count);
  m->infoset->members.Append(m);
  if (n->parent)
    n->parent->children[n->parent->children.Find(n)] = m;
  else
    root = m;
  m->children.Append(n);
  n->parent = m;
  while (--count)
    m->children.Append(new gbtEfgNode(this, m));

  DeleteLexicon();
  SortInfosets();
  return m->infoset;
}

gbtEfgInfoset *gbtEfgGame::InsertNode(gbtEfgNode *n, gbtEfgInfoset *s)
{
  if (!n || !s)  throw Exception();

  // can't bridge subgames
  if (s->members.Length() > 0 && n->gameroot != s->members[1]->gameroot)
    return 0;
  
  gbtEfgNode *m = new gbtEfgNode(this, n->parent);
  m->infoset = s;
  s->members.Append(m);
  if (n->parent)
    n->parent->children[n->parent->children.Find(n)] = m;
  else
    root = m;
  m->children.Append(n);
  n->parent = m;
  int count = s->actions.Length();
  while (--count)
    m->children.Append(new gbtEfgNode(this, m));

  DeleteLexicon();
  SortInfosets();
  return m->infoset;
}

gbtEfgInfoset *gbtEfgGame::CreateInfoset(gbtEfgPlayer *p, int br)
{
  if (!p || p->Game() != this)  throw Exception();

  return CreateInfoset(p->infosets.Length() + 1, p, br);
}

gbtEfgInfoset *gbtEfgGame::JoinInfoset(gbtEfgInfoset *s, gbtEfgNode *n)
{
  if (!n || !s)  throw Exception();

  // can't bridge subgames
  if (s->members.Length() > 0 && n->gameroot != s->members[1]->gameroot)
    return 0;
  
  if (!n->infoset)   return 0; 
  if (n->infoset == s)   return s;
  if (s->actions.Length() != n->children.Length())  return n->infoset;

  gbtEfgInfoset *t = n->infoset;

  t->members.Remove(t->members.Find(n));
  s->members.Append(n);

  n->infoset = s;

  DeleteLexicon();
  SortInfosets();
  return s;
}

gbtEfgInfoset *gbtEfgGame::LeaveInfoset(gbtEfgNode *n)
{
  if (!n)  throw Exception();

  if (!n->infoset)   return 0;

  gbtEfgInfoset *s = n->infoset;
  if (s->members.Length() == 1)   return s;

  gbtEfgPlayer *p = s->player;
  s->members.Remove(s->members.Find(n));
  n->infoset = CreateInfoset(p->infosets.Length() + 1, p,
			     n->children.Length());
  n->infoset->name = s->name;
  n->infoset->members.Append(n);
  for (int i = 1; i <= s->actions.Length(); i++)
    n->infoset->actions[i]->name = s->actions[i]->name;

  DeleteLexicon();
  SortInfosets();
  return n->infoset;
}

gbtEfgInfoset *gbtEfgGame::SplitInfoset(gbtEfgNode *n)
{
  if (!n)  throw Exception();

  if (!n->infoset)   return 0;

  gbtEfgInfoset *s = n->infoset;
  if (s->members.Length() == 1)   return s;

  gbtEfgPlayer *p = s->player;
  gbtEfgInfoset *ns = CreateInfoset(p->infosets.Length() + 1, p,
			      n->children.Length());
  ns->name = s->name;
  int i;
  for (i = s->members.Length(); i > s->members.Find(n); i--)   {
    gbtEfgNode *nn = s->members.Remove(i);
    ns->members.Append(nn);
    nn->infoset = ns;
  }
  for (i = 1; i <= s->actions.Length(); i++) {
    ns->actions[i]->name = s->actions[i]->name;
    if (p == chance) {
      SetChanceProb(ns, i, GetChanceProbText(s, i));
    }
  }
  DeleteLexicon();
  SortInfosets();
  return n->infoset;
}

gbtEfgInfoset *gbtEfgGame::MergeInfoset(gbtEfgInfoset *to, gbtEfgInfoset *from)
{
  if (!to || !from)  throw Exception();

  if (to == from ||
      to->actions.Length() != from->actions.Length())   return from;

  if (to->members[1]->gameroot != from->members[1]->gameroot) 
    return from;

  for (int i = 1; i <= from->members.Length(); i++) {
    to->members.Append(from->members[i]);
  }

  for (int i = 1; i <= from->members.Length(); i++)
    from->members[i]->infoset = to;

  from->members = gArray<gbtEfgNode *>();

  DeleteLexicon();
  SortInfosets();
  return to;
}

bool gbtEfgGame::DeleteEmptyInfoset(gbtEfgInfoset *s)
{
  if (!s)  throw Exception();

  if (s->NumMembers() > 0)   return false;

  s->player->infosets.Remove(s->player->infosets.Find(s));
  delete s;

  return true;
}

void gbtEfgGame::DeleteEmptyInfosets(void)
{
  for (int pl = 1; pl <= NumPlayers(); pl++) {
    for (int iset = 1; iset <= NumInfosets()[pl]; iset++) {
      if (DeleteEmptyInfoset(Players()[pl]->Infosets()[iset])) {
        iset--;
      }
    }
  }
} 

gbtEfgInfoset *gbtEfgGame::SwitchPlayer(gbtEfgInfoset *s, gbtEfgPlayer *p)
{
  if (!s || !p)  throw Exception();
  if (s->GetPlayer()->IsChance() || p->IsChance())  throw Exception();
  
  if (s->player == p)   return s;

  s->player->infosets.Remove(s->player->infosets.Find(s));
  s->player = p;
  p->infosets.Append(s);

  DeleteLexicon();
  SortInfosets();
  return s;
}

void gbtEfgGame::CopySubtree(gbtEfgNode *src, gbtEfgNode *dest, gbtEfgNode *stop)
{
  if (src == stop) {
    dest->outcome = src->outcome;
    return;
  }

  if (src->children.Length())  {
    AppendNode(dest, src->infoset);
    for (int i = 1; i <= src->children.Length(); i++)
      CopySubtree(src->children[i], dest->children[i], stop);
  }

  dest->m_label = src->m_label;
  dest->outcome = src->outcome;
}

//
// MarkSubtree: sets the Node::mark flag on all children of p_node
//
void gbtEfgGame::MarkSubtree(gbtEfgNode *p_node)
{
  p_node->mark = true;
  for (int i = 1; i <= p_node->children.Length(); i++) {
    MarkSubtree(p_node->children[i]);
  }
}

//
// UnmarkSubtree: clears the Node::mark flag on all children of p_node
//
void gbtEfgGame::UnmarkSubtree(gbtEfgNode *p_node)
{
  p_node->mark = false;
  for (int i = 1; i <= p_node->children.Length(); i++) {
    UnmarkSubtree(p_node->children[i]);
  }
}

void gbtEfgGame::Reveal(gbtEfgInfoset *where, const gArray<gbtEfgPlayer *> &who)
{
  UnmarkSubtree(root);  // start with a clean tree
  
  for (int i = 1; i <= where->actions.Length(); i++) {
    for (int j = 1; j <= where->members.Length(); j++) { 
      MarkSubtree(where->members[j]->children[i]);
    }

    for (int j = who.First(); j <= who.Last(); j++) {
      // iterate over each information set of player 'j' in the list
      for (int k = 1; k <= who[j]->infosets.Length(); k++) {
	// iterate over each member of information set 'k'
	// make copy of members to iterate correctly 
	// (since the information set may be changed in the process)
	gArray<gbtEfgNode *> members = who[j]->infosets[k]->members;
	gbtEfgInfoset *newiset = 0;

	for (int m = 1; m <= members.Length(); m++) {
	  gbtEfgNode *n = members[m];
	  if (n->mark) {
	    // If node is marked, is descendant of action 'i'
	    n->mark = false;   // unmark so tree is clean at end
	    if (!newiset) {
	      newiset = LeaveInfoset(n);
	    }
	    else {
	      JoinInfoset(newiset, n);
	    }
	  } 
	}
      }
    }
  }

  Reindex();
}

gbtEfgNode *gbtEfgGame::CopyTree(gbtEfgNode *src, gbtEfgNode *dest)
{
  if (!src || !dest)  throw Exception();
  if (src == dest || dest->children.Length())   return src;
  if (src->gameroot != dest->gameroot)  return src;

  if (src->children.Length())  {

    AppendNode(dest, src->infoset);
    for (int i = 1; i <= src->children.Length(); i++)
      CopySubtree(src->children[i], dest->children[i], dest);

    DeleteLexicon();
    SortInfosets();
  }

  return dest;
}

gbtEfgNode *gbtEfgGame::MoveTree(gbtEfgNode *src, gbtEfgNode *dest)
{
  if (!src || !dest)  throw Exception();
  if (src == dest || dest->children.Length() || IsPredecessor(src, dest))
    return src;
  if (src->gameroot != dest->gameroot)  return src;

  if (src->parent == dest->parent) {
    int srcChild = src->parent->children.Find(src);
    int destChild = src->parent->children.Find(dest);
    src->parent->children[srcChild] = dest;
    src->parent->children[destChild] = src;
  }
  else {
    gbtEfgNode *parent = src->parent; 
    parent->children[parent->children.Find(src)] = dest;
    dest->parent->children[dest->parent->children.Find(dest)] = src;
    src->parent = dest->parent;
    dest->parent = parent;
  }

  dest->m_label = "";
  dest->outcome = 0;
  
  DeleteLexicon();
  SortInfosets();
  return dest;
}

gbtEfgNode *gbtEfgGame::DeleteTree(gbtEfgNode *n)
{
  if (!n)  throw Exception();

  while (n->NumChildren() > 0)   {
    DeleteTree(n->children[1]);
    delete n->children.Remove(1);
  }
  
  if (n->infoset)  {
    n->infoset->members.Remove(n->infoset->members.Find(n));
    n->infoset = 0;
  }
  n->outcome = 0;
  n->m_label = "";

  DeleteLexicon();
  SortInfosets();
  return n;
}

gbtEfgAction *gbtEfgGame::InsertAction(gbtEfgInfoset *s)
{
  if (!s)  throw Exception();

  gbtEfgAction *action = s->InsertAction(s->NumActions() + 1);
  for (int i = 1; i <= s->members.Length(); i++) {
    s->members[i]->children.Append(new gbtEfgNode(this, s->members[i]));
  }
  DeleteLexicon();
  SortInfosets();
  return action;
}

gbtEfgAction *gbtEfgGame::InsertAction(gbtEfgInfoset *s, const gbtEfgAction *a)
{
  if (!a || !s)  throw Exception();

  int where;
  for (where = 1; where <= s->actions.Length() && s->actions[where] != a;
       where++);
  if (where > s->actions.Length())   return 0;
  gbtEfgAction *action = s->InsertAction(where);
  for (int i = 1; i <= s->members.Length(); i++)
    s->members[i]->children.Insert(new gbtEfgNode(this, s->members[i]), where);

  DeleteLexicon();
  SortInfosets();
  return action;
}

gbtEfgInfoset *gbtEfgGame::DeleteAction(gbtEfgInfoset *s, const gbtEfgAction *a)
{
  if (!a || !s)  throw Exception();

  int where;
  for (where = 1; where <= s->actions.Length() && s->actions[where] != a;
       where++);
  if (where > s->actions.Length() || s->actions.Length() == 1)   return s;
  s->RemoveAction(where);
  for (int i = 1; i <= s->members.Length(); i++)   {
    DeleteTree(s->members[i]->children[where]);
    delete s->members[i]->children.Remove(where);
  }
  DeleteLexicon();
  SortInfosets();
  return s;
}

void gbtEfgGame::SetChanceProb(gbtEfgInfoset *infoset, int act, 
			       const std::string &value)
{
  if (infoset->IsChanceInfoset()) {
    ((ChanceInfoset *) infoset)->SetActionProb(act, value);
    DeleteLexicon();
  }
}

gRational gbtEfgGame::GetChanceProb(gbtEfgInfoset *infoset, int act) const
{
  if (infoset->IsChanceInfoset())
    return ((ChanceInfoset *) infoset)->GetActionProb(act);
  else
    return gRational(0);
}

std::string gbtEfgGame::GetChanceProbText(gbtEfgInfoset *infoset,
					  int act) const
{
  if (infoset->IsChanceInfoset())
    return dynamic_cast<ChanceInfoset *>(infoset)->GetActionProbText(act);
  else
    return "";
}

gRational gbtEfgGame::GetChanceProb(const gbtEfgAction *a) const
{
  return GetChanceProbs(a->BelongsTo())[a->GetNumber()];
}

gArray<gRational> gbtEfgGame::GetChanceProbs(gbtEfgInfoset *infoset) const
{
  if (infoset->IsChanceInfoset())
    return ((ChanceInfoset *) infoset)->GetActionProbs();
  else
    return gArray<gRational>(infoset->NumActions());
}

//---------------------------------------------------------------------
//                     Subgame-related functions
//---------------------------------------------------------------------

void gbtEfgGame::MarkTree(gbtEfgNode *n, gbtEfgNode *base)
{
  n->ptr = base;
  for (int i = 1; i <= n->NumChildren(); i++)
    MarkTree(n->GetChild(i), base);
}

bool gbtEfgGame::CheckTree(gbtEfgNode *n, gbtEfgNode *base)
{
  int i;

  if (n->NumChildren() == 0)   return true;

  for (i = 1; i <= n->NumChildren(); i++)
    if (!CheckTree(n->GetChild(i), base))  return false;

  if (n->GetPlayer()->IsChance())   return true;

  for (i = 1; i <= n->GetInfoset()->NumMembers(); i++)
    if (n->GetInfoset()->Members()[i]->ptr != base)
      return false;

  return true;
}

bool gbtEfgGame::IsLegalSubgame(gbtEfgNode *n)
{
  if (n->NumChildren() == 0)  
    return false;

  MarkTree(n, n);
  return CheckTree(n, n);
}

bool gbtEfgGame::MarkSubgame(gbtEfgNode *n)
{
  if(n->gameroot == n) return true;

  if (n->gameroot != n && IsLegalSubgame(n))  {
    n->gameroot = 0;
    MarkSubgame(n, n);
    return true;
  }

  return false;
}

void gbtEfgGame::UnmarkSubgame(gbtEfgNode *n)
{
  if (n->gameroot == n && n->parent)  {
    n->gameroot = 0;
    MarkSubgame(n, n->parent->gameroot);
  }
}
  

void gbtEfgGame::MarkSubgame(gbtEfgNode *n, gbtEfgNode *base)
{
  if (n->gameroot == n)  return;
  n->gameroot = base;
  for (int i = 1; i <= n->NumChildren(); i++)
    MarkSubgame(n->GetChild(i), base);
}

void gbtEfgGame::MarkSubgames(const gList<gbtEfgNode *> &list)
{
  for (int i = 1; i <= list.Length(); i++)  {
    if (IsLegalSubgame(list[i]))  {
      list[i]->gameroot = 0;
      MarkSubgame(list[i], list[i]);
    }
  }
}

void gbtEfgGame::MarkSubgames(void)
{
  gList<gbtEfgNode *> subgames;
  LegalSubgameRoots(*this, subgames);

  for (int i = 1; i <= subgames.Length(); i++)  {
    subgames[i]->gameroot = 0;
    MarkSubgame(subgames[i], subgames[i]);
  }
}

void gbtEfgGame::UnmarkSubgames(gbtEfgNode *n)
{
  if (n->NumChildren() == 0)   return;

  for (int i = 1; i <= n->NumChildren(); i++)
    UnmarkSubgames(n->GetChild(i));
  
  if (n->gameroot == n && n->parent)  {
    n->gameroot = 0;
    MarkSubgame(n, n->parent->gameroot);
  }
}


int gbtEfgGame::ProfileLength(void) const
{
  int sum = 0;

  for (int i = 1; i <= players.Length(); i++)
    for (int j = 1; j <= players[i]->infosets.Length(); j++)
      sum += players[i]->infosets[j]->actions.Length();

  return sum;
}

gArray<int> gbtEfgGame::NumInfosets(void) const
{
  gArray<int> foo(players.Length());
  
  for (int i = 1; i <= foo.Length(); i++) {
    foo[i] = players[i]->NumInfosets();
  }

  return foo;
}

int gbtEfgGame::NumPlayerInfosets() const
{
  int answer(0);
  int pl;
  for (pl = 1; pl <= NumPlayers(); pl++)
    answer +=  players[pl]->infosets.Length();
  return answer;
}

int gbtEfgGame::NumChanceInfosets() const
{
  return chance->infosets.Length();
}

int gbtEfgGame::TotalNumInfosets(void) const
{
  return NumPlayerInfosets() + NumChanceInfosets();
}

gPVector<int> gbtEfgGame::NumActions(void) const
{
  gArray<int> foo(players.Length());
  int i;
  for (i = 1; i <= players.Length(); i++)
    foo[i] = players[i]->infosets.Length();

  gPVector<int> bar(foo);
  for (i = 1; i <= players.Length(); i++) {
    for (int j = 1; j <= players[i]->infosets.Length(); j++) {
      bar(i, j) = players[i]->infosets[j]->NumActions();
    }
  }

  return bar;
}  

int gbtEfgGame::NumPlayerActions(void) const
{
  int answer = 0;

  gPVector<int> nums_actions = NumActions();
  for (int i = 1; i <= NumPlayers(); i++)
    answer += nums_actions[i];
  return answer;
}

int gbtEfgGame::NumChanceActions(void) const
{
  int answer = 0;

  for (int i = 1; i <= players[0]->infosets.Length(); i++) {
    answer += players[0]->infosets[i]->NumActions();
  }

  return answer;
}

gPVector<int> gbtEfgGame::NumMembers(void) const
{
  gArray<int> foo(players.Length());

  for (int i = 1; i <= players.Length(); i++) {
    foo[i] = players[i]->NumInfosets();
  }

  gPVector<int> bar(foo);
  for (int i = 1; i <= players.Length(); i++) {
    for (int j = 1; j <= players[i]->NumInfosets(); j++) {
      bar(i, j) = players[i]->infosets[j]->NumMembers();
    }
  }

  return bar;
}

//------------------------------------------------------------------------
//                       Efg: Payoff computation
//------------------------------------------------------------------------

void gbtEfgGame::Payoff(gbtEfgNode *n, gRational prob, 
			const gPVector<int> &profile,
			gVector<gRational> &payoff) const
{
  if (n->outcome)  {
    for (int i = 1; i <= players.Length(); i++)
      payoff[i] += prob * n->outcome->m_ratPayoffs[i];
  }

  if (n->infoset && n->infoset->player->IsChance())
    for (int i = 1; i <= n->children.Length(); i++)
      Payoff(n->children[i],
	     prob * GetChanceProb(n->infoset, i),
	     profile, payoff);
  else if (n->infoset)
    Payoff(n->children[profile(n->infoset->player->number,n->infoset->number)],
	   prob, profile, payoff);
}

void gbtEfgGame::InfosetProbs(gbtEfgNode *n, gRational prob, 
			      const gPVector<int> &profile,
			      gPVector<gRational> &probs) const
{
  if (n->infoset && n->infoset->player->IsChance())
    for (int i = 1; i <= n->children.Length(); i++)
      InfosetProbs(n->children[i],
		   prob * GetChanceProb(n->infoset, i),
		   profile, probs);
  else if (n->infoset)  {
    probs(n->infoset->player->number, n->infoset->number) += prob;
    InfosetProbs(n->children[profile(n->infoset->player->number,n->infoset->number)],
		 prob, profile, probs);
  }
}

void gbtEfgGame::Payoff(const gPVector<int> &profile, 
			gVector<gRational> &payoff) const
{
  ((gVector<gRational> &) payoff).operator=(gRational(0));
  Payoff(root, 1, profile, payoff);
}

void gbtEfgGame::InfosetProbs(const gPVector<int> &profile,
			      gPVector<gRational> &probs) const
{
  ((gVector<gRational> &) probs).operator=(gRational(0));
  InfosetProbs(root, 1, profile, probs);
}

void gbtEfgGame::Payoff(gbtEfgNode *n, gRational prob, 
			const gArray<gArray<int> *> &profile,
			gArray<gRational> &payoff) const
{
  if (n->outcome)   {
    for (int i = 1; i <= players.Length(); i++)
      payoff[i] += prob * n->outcome->m_ratPayoffs[i];
  }
  
  if (n->infoset && n->infoset->player->IsChance())
    for (int i = 1; i <= n->children.Length(); i++)
      Payoff(n->children[i],
	     prob * GetChanceProb(n->infoset, i),
	     profile, payoff);
  else if (n->infoset)
    Payoff(n->children[(*profile[n->infoset->player->number])[n->infoset->number]],
	   prob, profile, payoff);
}

void gbtEfgGame::Payoff(const gArray<gArray<int> *> &profile,
			gArray<gRational> &payoff) const
{
  for (int i = 1; i <= payoff.Length(); i++)
    payoff[i] = 0;
  Payoff(root, 1, profile, payoff);
}

gbtNfgGame *gbtEfgGame::AssociatedNfg(void) const
{
  if (lexicon) {
    return lexicon->N;
  }
  else {
    return 0;
  }
}


