//
// $Source: /cvsroot/gambit/gambit/sources/game/Attic/efg.h,v $
// $Date: 2005/06/13 02:32:09 $
// $Revision: 1.6.2.11 $
//
// DESCRIPTION:
// Interface to extensive form representation class
//
// This file is part of Gambit
// Copyright (c) 2002, The Gambit Project
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
//

#ifndef EFG_H
#define EFG_H

#include "base/base.h"
#include "math/rational.h"
#include "math/gpvector.h"

class gbtEfgPlayer;
class gbtEfgInfoset;
class gbtEfgNode;
class gbtEfgAction;
class gbtEfgSupport;
class Lexicon;
class gbtNfgGame;
class gNumber;
template <class T> class BehavProfile;
template <class T> class MixedProfile;
template <class T> class PureBehavProfile;

class gbtEfgOutcome;

class gbtEfgGame {
private:
  friend class EfgFileReader;
  friend class EfgFile;
  friend class gbtNfgGame;
  friend class gbtEfgNode;
  friend class BehavProfile<double>;
  friend class BehavProfile<gRational>;
  friend class BehavProfile<gNumber>;
  
protected:
  bool sortisets;
  std::string title, comment;
  gArray<gbtEfgPlayer *> players;
  gArray<gbtEfgOutcome *> outcomes;
  gbtEfgNode *root;
  gbtEfgPlayer *chance;
  mutable Lexicon *lexicon;
  
  // this is for use with the copy constructor
  void CopySubtree(gbtEfgNode *, gbtEfgNode *);

  void CopySubtree(gbtEfgNode *, gbtEfgNode *, gbtEfgNode *);
  void MarkSubtree(gbtEfgNode *);
  void UnmarkSubtree(gbtEfgNode *);

  void SortInfosets(void);
  void NumberNodes(gbtEfgNode *, int &);
  
  void DeleteLexicon(void) const;

  gbtEfgOutcome *NewOutcome(int index);

  void WriteEfgFile(std::ostream &, gbtEfgNode *) const;

  void Payoff(gbtEfgNode *n, gRational, const gPVector<int> &, gVector<gRational> &) const;
  void Payoff(gbtEfgNode *n, gRational, const gArray<gArray<int> *> &, gArray<gRational> &) const;
  
  void InfosetProbs(gbtEfgNode *n, gRational, const gPVector<int> &, gPVector<gRational> &) const;
    
    
  // These are used in identification of subgames
  void MarkTree(gbtEfgNode *, gbtEfgNode *);
  bool CheckTree(gbtEfgNode *, gbtEfgNode *);
  void MarkSubgame(gbtEfgNode *, gbtEfgNode *);

  // Recursive calls
  void DescendantNodes(const gbtEfgNode *, const gbtEfgSupport &, 
		       gList<gbtEfgNode *> &) const;
  void NonterminalDescendants(const gbtEfgNode *, const gbtEfgSupport&, 
			      gList<const gbtEfgNode*> &) const;
  void TerminalDescendants(const gbtEfgNode *, const gbtEfgSupport&, 
			   gList<gbtEfgNode *> &) const;

public:
  class Exception : public gException   {
  public:
    virtual ~Exception()   { }
    std::string Description(void) const    { return "Efg error"; }
  };

  gbtEfgGame(void);
  gbtEfgGame(const gbtEfgGame &, gbtEfgNode * = 0);
  virtual ~gbtEfgGame();
  
  // TITLE ACCESS AND MANIPULATION
  void SetTitle(const std::string &s);
  const std::string &GetTitle(void) const;
  
  void SetComment(const std::string &);
  const std::string &GetComment(void) const;

  // WRITING DATA FILES
  void WriteEfgFile(std::ostream &p_file) const;

  // DATA ACCESS -- GENERAL INFORMATION
  bool IsConstSum(void) const; 
  gRational MinPayoff(int pl = 0) const;
  gRational MaxPayoff(int pl = 0) const;
  
  gbtEfgNode *RootNode(void) const;
  bool IsSuccessor(const gbtEfgNode *n, const gbtEfgNode *from) const;
  bool IsPredecessor(const gbtEfgNode *n, const gbtEfgNode *of) const;
  gArray<int> PathToNode(const gbtEfgNode *) const;
  //    const Node* Consequence(const Node&, Action&) const;
  gList<gbtEfgNode *> DescendantNodes(const gbtEfgNode&, const gbtEfgSupport&) const;
  gList<const gbtEfgNode*> NonterminalDescendants(const gbtEfgNode&, 
						  const gbtEfgSupport&) const;
  gList<gbtEfgNode *> TerminalDescendants(const gbtEfgNode&, 
				    const gbtEfgSupport&) const;
  gList<gbtEfgNode *> TerminalNodes(void) const;
  gList<gbtEfgInfoset*> DescendantInfosets(const gbtEfgNode&, const gbtEfgSupport&) const;

  // DATA ACCESS -- PLAYERS
  int NumPlayers(void) const;
  gbtEfgPlayer *GetPlayer(int pl) const { return players[pl]; }
  gbtEfgPlayer *GetChance(void) const;
  gbtEfgPlayer *NewPlayer(void);
  const gArray<gbtEfgPlayer *> &Players(void) const  { return players; }

  // DATA ACCESS -- INFOSETS
  gArray<gbtEfgInfoset *> Infosets(void) const;

  // DATA ACCESS -- OUTCOMES
  int NumOutcomes(void) const;
  gbtEfgOutcome *GetOutcome(int index) const;
  gbtEfgOutcome *NewOutcome(void);
  void DeleteOutcome(gbtEfgOutcome *);

  // EDITING OPERATIONS
  gbtEfgInfoset *AppendNode(gbtEfgNode *n, gbtEfgPlayer *p, int br);
  gbtEfgInfoset *AppendNode(gbtEfgNode *n, gbtEfgInfoset *s);
  gbtEfgNode *DeleteNode(gbtEfgNode *n, gbtEfgNode *keep);
  gbtEfgInfoset *InsertNode(gbtEfgNode *n, gbtEfgPlayer *p, int br);
  gbtEfgInfoset *InsertNode(gbtEfgNode *n, gbtEfgInfoset *s);

  gbtEfgInfoset *CreateInfoset(gbtEfgPlayer *pl, int br);
  bool DeleteEmptyInfoset(gbtEfgInfoset *);
  void DeleteEmptyInfosets(void);
  gbtEfgInfoset *JoinInfoset(gbtEfgInfoset *s, gbtEfgNode *n);
  gbtEfgInfoset *LeaveInfoset(gbtEfgNode *n);
  gbtEfgInfoset *SplitInfoset(gbtEfgNode *n);
  gbtEfgInfoset *MergeInfoset(gbtEfgInfoset *to, gbtEfgInfoset *from);

  gbtEfgInfoset *SwitchPlayer(gbtEfgInfoset *s, gbtEfgPlayer *p);
  
  gbtEfgNode *CopyTree(gbtEfgNode *src, gbtEfgNode *dest);
  gbtEfgNode *MoveTree(gbtEfgNode *src, gbtEfgNode *dest);
  gbtEfgNode *DeleteTree(gbtEfgNode *n);

  gbtEfgAction *InsertAction(gbtEfgInfoset *s);
  gbtEfgAction *InsertAction(gbtEfgInfoset *s, const gbtEfgAction *at);
  gbtEfgInfoset *DeleteAction(gbtEfgInfoset *s, const gbtEfgAction *a);

  void Reveal(gbtEfgInfoset *, const gArray<gbtEfgPlayer *> &);

  void SetChanceProb(gbtEfgInfoset *, int, const std::string &);
  gRational GetChanceProb(gbtEfgInfoset *, int) const;
  std::string GetChanceProbText(gbtEfgInfoset *, int) const;
  gRational GetChanceProb(const gbtEfgAction *) const;
  gArray<gRational> GetChanceProbs(gbtEfgInfoset *) const;

  void SetPayoff(gbtEfgOutcome *, int pl, const std::string &value);
  gRational Payoff(gbtEfgOutcome *, const gbtEfgPlayer *) const;
  std::string GetPayoffText(gbtEfgOutcome *, const gbtEfgPlayer *) const;

  gRational Payoff(const gbtEfgNode *, const gbtEfgPlayer *) const;
  gArray<gRational> Payoff(gbtEfgOutcome *) const;

  bool IsLegalSubgame(gbtEfgNode *n);
  void MarkSubgames(void);
  void MarkSubgames(const gList<gbtEfgNode *> &list);
  bool MarkSubgame(gbtEfgNode *n);
  void UnmarkSubgame(gbtEfgNode *n);
  void UnmarkSubgames(gbtEfgNode *n);

  int ProfileLength(void) const;
  int TotalNumInfosets(void) const;

  gArray<int>   NumInfosets(void) const;  // Does not include chance infosets
  int           NumPlayerInfosets(void) const;
  int           NumChanceInfosets(void) const;
  gPVector<int> NumActions(void) const;
  int           NumPlayerActions(void) const;
  int           NumChanceActions(void) const;
  gPVector<int> NumMembers(void) const;
  
  // COMPUTING VALUES OF PROFILES
  void Payoff(const gPVector<int> &profile, gVector<gRational> &payoff) const;
  void Payoff(const gArray<gArray<int> *> &profile,
	      gArray<gRational> &payoff) const;

  void InfosetProbs(const gPVector<int> &profile, gPVector<gRational> &prob) const;
    
  gbtNfgGame *AssociatedNfg(void) const;
  Lexicon *GetLexicon(void) const { return lexicon; }

  friend gbtNfgGame *MakeReducedNfg(const gbtEfgSupport &);

  // These are auxiliary functions used by the .efg file reader code
  gbtEfgInfoset *GetInfosetByIndex(gbtEfgPlayer *p, int index) const;
  gbtEfgInfoset *CreateInfosetByIndex(gbtEfgPlayer *p, int index, int br);
  gbtEfgOutcome *GetOutcomeByIndex(int index) const;
  gbtEfgOutcome *CreateOutcomeByIndex(int index);
  void Reindex(void);
  gbtEfgInfoset *CreateInfoset(int n, gbtEfgPlayer *pl, int br);
};

//#include "behav.h"

#include "efplayer.h"
#include "infoset.h"
#include "node.h"

class gbtEfgOutcome   {
friend class gbtEfgGame;
friend class BehavProfile<double>;
friend class BehavProfile<gRational>;
friend class BehavProfile<gNumber>;
protected:
  gbtEfgGame *m_efg; 
  int m_number;
  std::string m_label;
  gArray<std::string> m_textPayoffs;
  gArray<gRational> m_ratPayoffs;
  gArray<double> m_doublePayoffs;

  gbtEfgOutcome(gbtEfgGame *p_efg, int p_number)
    : m_efg(p_efg), m_number(p_number), 
      m_textPayoffs(p_efg->NumPlayers()), 
      m_ratPayoffs(p_efg->NumPlayers()),
      m_doublePayoffs(p_efg->NumPlayers())
    {
      for (int pl = 1; pl <= m_textPayoffs.Length(); pl++) {
	m_textPayoffs[pl] = "0";
	m_doublePayoffs[pl] = 0.0;
      }
    }
  ~gbtEfgOutcome()  { }

public:
  gbtEfgGame *GetGame(void) const { return m_efg; }

  const std::string &GetLabel(void) const { return m_label; }
  void SetLabel(const std::string &p_label) { m_label = p_label; }
};

gbtEfgGame *ReadEfgFile(std::istream &);

template <class T> class PureBehavProfile   {
  protected:
    const gbtEfgGame *E;
    gArray<gArray<const gbtEfgAction *> *> profile;

    //    void IndPayoff(const gbtEfgNode *n, const int &pl, const T, T &) const;
    // This aims at efficiency, but leads to a problem described in behav.imp

    void Payoff(const gbtEfgNode *n, const T, gArray<T> &) const;
    void InfosetProbs(gbtEfgNode *n, T, gPVector<T> &) const;

  public:
    PureBehavProfile(const gbtEfgGame &);
    PureBehavProfile(const PureBehavProfile<T> &);
    ~PureBehavProfile();

    // Operators
    PureBehavProfile<T> &operator=(const PureBehavProfile<T> &);
    T operator()(gbtEfgAction *) const;

    // Manipulation
    void Set(const gbtEfgAction *);
    void Set(const gbtEfgPlayer *, const gArray<const gbtEfgAction *> &);

    // Information
    const gbtEfgAction *GetAction(const gbtEfgInfoset *) const;
    
    const T Payoff(gbtEfgOutcome *, const int &pl) const;
    const T ChanceProb(const gbtEfgInfoset *, const int &act) const;
    
    const T Payoff(const gbtEfgNode *, const int &pl) const;
  //    T    Payoff(const int &pl) const;
    void Payoff(gArray<T> &payoff) const;
    void InfosetProbs(gPVector<T> &prob) const;
    gbtEfgGame &GetGame(void) const   { return const_cast<gbtEfgGame &>(*E); }
};


#include "efgutils.h"

#endif   // EFG_H


