//
// $Source: /cvsroot/gambit/gambit/sources/game/Attic/efgutils.cc,v $
// $Date: 2005/06/09 01:50:51 $
// $Revision: 1.4.2.5 $
//
// DESCRIPTION:
// Utility functions on extensive form games
//
// This file is part of Gambit
// Copyright (c) 2002, The Gambit Project
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
//

#include "efgutils.h"
#include "efstrat.h"

// recursive functions

static void NDoChild(const gbtEfgGame &e, gbtEfgNode *n, gList <gbtEfgNode *> &list)
{ 
  list.Append(n);
  for (int i = 1; i <= n->NumChildren(); i++)
    NDoChild (e, n->GetChild(i), list);
}

static void MSRDoChild(const gbtEfgGame &e, gbtEfgNode *n, gList<gbtEfgNode *> &list)
{
  for (int i = 1; i <= n->NumChildren(); i++)
    MSRDoChild(e, n->GetChild(i), list);
  if (n->GetSubgameRoot() == n)  list.Append(n);
}

static void LSRDoChild(const gbtEfgGame &e, gbtEfgNode *n, gList<gbtEfgNode *> &list)
{
  for (int i = 1; i <= n->NumChildren(); i++)
    LSRDoChild(e, n->GetChild(i), list);
  if (n->Game()->IsLegalSubgame(n))   list.Append(n);
}

static void CSDoChild(const gbtEfgGame &e, gbtEfgNode *n, gList<gbtEfgNode *> &list)
{
  if (n->GetSubgameRoot() == n)
    list.Append(n);
  else
    for (int i = 1; i <= n->NumChildren(); i++)
      CSDoChild(e, n->GetChild(i), list);
}

// Public Functions
 
int CountNodes (const gbtEfgGame &e, gbtEfgNode *n)
{
  int num = 1;
  for (int i = 1; i <= n->NumChildren(); i++)
    num += CountNodes (e, n->GetChild(i));
  return num;
}

void Nodes (const gbtEfgGame &befg, gList <gbtEfgNode *> &list)
{
  list = gList<gbtEfgNode *>();
  NDoChild(befg, befg.RootNode(), list); 
}

void Nodes (const gbtEfgGame &efg, gbtEfgNode *n, gList <gbtEfgNode *> &list)
{
  list = gList<gbtEfgNode *>();
  NDoChild(efg,n, list);
}

void MarkedSubgameRoots(const gbtEfgGame &efg, gList<gbtEfgNode *> &list)
{
  list = gList<gbtEfgNode *>();
  MSRDoChild(efg, efg.RootNode(), list);
}

void LegalSubgameRoots(const gbtEfgGame &efg, gList<gbtEfgNode *> &list)
{
  list = gList<gbtEfgNode *>();
  LSRDoChild(efg, efg.RootNode(), list);
}

void LegalSubgameRoots(const gbtEfgGame &efg, gbtEfgNode *n, gList<gbtEfgNode *> &list)
{
  list = gList<gbtEfgNode *>();
  LSRDoChild(efg, n, list);
}

bool HasSubgames(const gbtEfgGame &efg)
{
  gList<gbtEfgNode *> list;
  LegalSubgameRoots(efg, list);
  return list.Length()>1;
}

bool HasSubgames(const gbtEfgGame &e, gbtEfgNode * n)
{
  gList<gbtEfgNode *> list;
  LegalSubgameRoots(e, n, list);
  if(n->Game()->IsLegalSubgame(n))
    return list.Length()>1;
  return list.Length()>0;
}

bool AllSubgamesMarked(const gbtEfgGame &efg)
{
  gList<gbtEfgNode *> marked, valid;

  LegalSubgameRoots(efg, valid);
  MarkedSubgameRoots(efg, marked);

  return (marked == valid);
}


void ChildSubgames(const gbtEfgGame &efg, gbtEfgNode *n, gList<gbtEfgNode *> &list)
{
  list = gList<gbtEfgNode *>();
  for (int i = 1; i <= n->NumChildren(); i++)
    CSDoChild(efg, n->GetChild(i), list);
}

int NumNodes (const gbtEfgGame &befg)
{
  return (CountNodes(befg, befg.RootNode()));
}

gbtEfgAction *LastAction(const gbtEfgGame &e, gbtEfgNode *node)
{
  gbtEfgNode *parent = node->GetParent();
  if (parent == 0)  return 0;
  for (int i = 1; i <= parent->NumChildren(); i++) 
    if (parent->GetChild(i) == node)  
      return parent->GetInfoset()->Actions()[i];
  return 0;
}

bool IsPerfectRecall(const gbtEfgGame &p_efg)
{
  gbtEfgInfoset *s1, *s2;
  return IsPerfectRecall(p_efg, s1, s2);
}

bool IsPerfectRecall(const gbtEfgGame &efg, gbtEfgInfoset *&s1, gbtEfgInfoset *&s2)
{
  for (int pl = 1; pl <= efg.NumPlayers(); pl++)   {
    gbtEfgPlayer *player = efg.Players()[pl];
    
    for (int i = 1; i <= player->NumInfosets(); i++)  {
      gbtEfgInfoset *iset1 = player->Infosets()[i];
      for (int j = 1; j <= player->NumInfosets(); j++)   {
	gbtEfgInfoset *iset2 = player->Infosets()[j];

	bool precedes = false;
	int action = 0;
	
	for (int m = 1; m <= iset2->NumMembers(); m++)  {
	  int n;
	  for (n = 1; n <= iset1->NumMembers(); n++)  {
	    if (efg.IsPredecessor(iset1->Members()[n],
				  iset2->Members()[m]) &&
	        iset1->Members()[n] != iset2->Members()[m])  {
	      precedes = true;
	      for (int act = 1; act <= iset1->NumActions(); act++)  {
		if (efg.IsPredecessor(iset1->Members()[n]->GetChild(act),
				      iset2->Members()[m]))  {
		  if (action != 0 && action != act)  {
		    s1 = iset1;
		    s2 = iset2;
		    return false;
		  }
		  action = act;
		}
	      }
	      break;
	    }
	  }
	  
	  if (i == j && precedes)  {
	    s1 = iset1;
	    s2 = iset2;
	    return false;
	  }

	  if (n > iset1->NumMembers() && precedes)  {
	    s1 = iset1;
	    s2 = iset2;
	    return false;
	  }
	}
	

      }
    }
  }

  return true;
}

