//
// $Source: /cvsroot/gambit/gambit/sources/game/Attic/nfg.cc,v $
// $Date: 2005/06/13 02:32:10 $
// $Revision: 1.3.2.10 $
//
// DESCRIPTION:
// Implementation of normal form game classes
//
// This file is part of Gambit
// Copyright (c) 2002, The Gambit Project
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
//

#include <iostream>

#include "math/rational.h"
#include "nfg.h"
#include "nfstrat.h"
#include "nfplayer.h"

static std::string EscapeQuotes(const std::string &s)
{
  std::string ret;
  
  for (unsigned int i = 0; i < s.length(); i++)  {
    if (s[i] == '"')   ret += '\\';
    ret += s[i];
  }

  return ret;
}

//--------------------------------------
// Strategy:  Constructors, Destructors
//--------------------------------------

Strategy::Strategy(gbtNfgPlayer *p) : m_number(0), m_player(p), m_index(0L)
{ }

Strategy::Strategy(const Strategy &s) : m_player(s.m_player), m_name(s.m_name)
{ }

Strategy &Strategy::operator=(const Strategy &s)
{
  m_player = s.m_player;
  m_name = s.m_name;
  return *this;
}

Strategy::~Strategy()
{ }


//----------------------------------------------------
// Nfg: Constructors, Destructors, Operators
//----------------------------------------------------


int gbtNfgGame::Product(const gArray<int> &dim)
{
  int accum = 1;
  for (int i = 1; i <= dim.Length(); accum *= dim[i++]);
  return accum;
}
  
gbtNfgGame::gbtNfgGame(const gArray<int> &dim)
  : title("Untitled normal form game"), dimensions(dim), players(dim.Length()),
    results(Product(dim)), efg(0)
{
  for (int pl = 1; pl <= players.Length(); pl++)  {
    players[pl] = new gbtNfgPlayer(pl, this, dim[pl]);
	  players[pl]->name = ToText(pl);
    for (int st = 1; st <= players[pl]->NumStrats(); st++)
      players[pl]->strategies[st]->m_name = ToText(st);
  }
  IndexStrategies();

  for (int cont = 1; cont <= results.Length();
       results[cont++] = (gbtNfgOutcome *) 0);
}

gbtNfgGame::gbtNfgGame(const gbtNfgGame &b)
  : title(b.title), comment(b.comment), 
    dimensions(b.dimensions),
    players(b.players.Length()), outcomes(b.outcomes.Length()),
    results(b.results.Length()), efg(0)
{
  for (int pl = 1; pl <= players.Length(); pl++)  {
    players[pl] = new gbtNfgPlayer(pl, this, dimensions[pl]);
    players[pl]->name = b.players[pl]->name;
    for (int st = 1; st <= players[pl]->NumStrats(); st++)  {
      *(players[pl]->strategies[st]) = *(b.players[pl]->strategies[st]);
      players[pl]->strategies[st]->m_player = players[pl];
    }
  }
  IndexStrategies();
  
  for (int outc = 1; outc <= outcomes.Length(); outc++)  {
    outcomes[outc] = new gbtNfgOutcome(outc, this);
    outcomes[outc]->SetName(b.outcomes[outc]->GetName());
    outcomes[outc]->m_textPayoffs = b.outcomes[outc]->m_textPayoffs;
    outcomes[outc]->m_ratPayoffs = b.outcomes[outc]->m_ratPayoffs;
    outcomes[outc]->m_doublePayoffs = b.outcomes[outc]->m_doublePayoffs;
  }

  for (int cont = 1; cont <= results.Length(); cont++)    
    results[cont] = (b.results[cont]) ?
                     outcomes[b.results[cont]->GetNumber()] : (gbtNfgOutcome *) 0;
}

#include "efg.h"
#include "lexicon.h"

gbtNfgGame::~gbtNfgGame()
{
  for (int pl = 1; pl <= players.Length(); pl++)
    delete players[pl];
  for (int outc = 1; outc <= outcomes.Length(); outc++)
    delete outcomes[outc];

  if (efg)  {
    const gbtEfgGame *tmp = (gbtEfgGame *) efg;
    // note that Lexicon dtor unsets the efg member...

    delete ((gbtEfgGame *) efg)->lexicon;
    tmp->lexicon = 0;
  }
  efg = 0;
}

void gbtNfgGame::BreakLink(void)
{
  if (efg)  {
    const gbtEfgGame *tmp = (gbtEfgGame *) efg;
    // note that Lexicon dtor unsets the efg member...

    delete ((gbtEfgGame *) efg)->lexicon;
    tmp->lexicon = 0;
  }
  efg = 0;
}

//-------------------------------
// Nfg: Member Functions
//-------------------------------

#include "nfgiter.h"
#include "nfgciter.h"

void gbtNfgGame::WriteNfgFile(std::ostream &p_file) const
{ 
  p_file << "NFG 1 R";
  p_file << " \"" << EscapeQuotes(GetTitle()) << "\" { ";

  for (int i = 1; i <= NumPlayers(); i++)
    p_file << '"' << EscapeQuotes(Players()[i]->GetName()) << "\" ";

  p_file << "}\n\n{ ";
  
  for (int i = 1; i <= NumPlayers(); i++)   {
    gbtNfgPlayer *player = Players()[i];
    p_file << "{ ";
    for (int j = 1; j <= player->NumStrats(); j++)
      p_file << '"' << EscapeQuotes(player->Strategies()[j]->Name()) << "\" ";
    p_file << "}\n";
  }
  
  p_file << "}\n";

  p_file << "\"" << EscapeQuotes(comment) << "\"\n\n";

  int ncont = 1;
  for (int i = 1; i <= NumPlayers(); i++)
    ncont *= NumStrats(i);

  p_file << "{\n";
  for (int outc = 1; outc <= outcomes.Length(); outc++)   {
    p_file << "{ \"" << EscapeQuotes(outcomes[outc]->name) << "\" ";
    for (int pl = 1; pl <= players.Length(); pl++)  {
      p_file << outcomes[outc]->m_textPayoffs[pl];

      if (pl < players.Length())
	p_file << ", ";
      else
	p_file << " }\n";
    }
  }
  p_file << "}\n";
  
  for (int cont = 1; cont <= ncont; cont++)  {
    if (results[cont] != 0)
      p_file << results[cont]->number << ' ';
    else
      p_file << "0 ";
  }

  p_file << '\n';
}

void gbtNfgGame::WriteHtmlFile(std::ostream &p_file, 
			int p_rowPlayer, int p_colPlayer) const
{
  std::string theHtml;
  gbtNfgSupport support(*this);
  NfgContIter iter(support);
  iter.Freeze(p_rowPlayer);
  iter.Freeze(p_colPlayer);
  iter.First();

  theHtml += "<center><h1>" + GetTitle() + "</h1></center>\n";

  do {
    StrategyProfile profile(iter.Profile());

    if (NumPlayers() > 2) {
      theHtml += "<center><b>Subtable with strategies:</b></center>";
      for (int pl = 1; pl <= NumPlayers(); pl++) {
	if (pl == p_rowPlayer || pl == p_colPlayer) {
	  continue;
	}

	theHtml += "<center><b>Player ";
	theHtml += ToText(pl);
	theHtml += " Strategy ";
	theHtml += ToText(profile.Get(pl)->Number());
	theHtml += "</b></center>";
      }
    }
    
    theHtml += "<table>";
    theHtml += "<tr>";
    theHtml += "<td></td>";
    for (int st = 1; st <= NumStrats(p_colPlayer); st++) {
      theHtml += "<td align=center><b>";
      theHtml += GetPlayer(p_colPlayer)->GetStrategy(st)->Name();
      theHtml += "</b></td>";
    } 
    theHtml += "</tr>";
    for (int st1 = 1; st1 <= NumStrats(p_rowPlayer); st1++) {
      profile.Set(p_rowPlayer, GetPlayer(p_rowPlayer)->GetStrategy(st1));
      theHtml += "<tr>";
      theHtml += "<td align=center><b>";
      theHtml += GetPlayer(p_rowPlayer)->GetStrategy(st1)->Name();
      theHtml += "</b></td>";
      for (int st2 = 1; st2 <= NumStrats(p_colPlayer); st2++) {
	profile.Set(p_colPlayer, GetPlayer(p_colPlayer)->GetStrategy(st2));
	theHtml += "<td align=center>";
	for (int pl = 1; pl <= NumPlayers(); pl++) {
	  theHtml += ToText(Payoff(GetOutcome(profile.GetIndex()+1), pl));
	  if (pl < NumPlayers()) {
	    theHtml += ",";
	  }
	}
	theHtml += "</td>";
      }
      theHtml += "</tr>";
    } 

    theHtml += "</table>";
  } while (iter.NextContingency());

  p_file << theHtml;
}

void gbtNfgGame::WriteOsborneFile(std::ostream &p_file, 
			   int p_rowPlayer, int p_colPlayer) const
{
  std::string theHtml;
  gbtNfgSupport support(*this);
  NfgContIter iter(support);
  iter.Freeze(p_rowPlayer);
  iter.Freeze(p_colPlayer);
  iter.First();

  do {
    StrategyProfile profile(iter.Profile());

    theHtml += "\\begin{game}{";
    theHtml += ToText(NumStrats(p_rowPlayer));
    theHtml += "}{";
    theHtml += ToText(NumStrats(p_colPlayer));
    theHtml += "}[";
    theHtml += GetPlayer(p_rowPlayer)->GetName();
    theHtml += "][";
    theHtml += GetPlayer(p_colPlayer)->GetName();
    theHtml += "]";

    if (NumPlayers() > 2) {
      theHtml += "[";
      for (int pl = 1; pl <= NumPlayers(); pl++) {
	if (pl == p_rowPlayer || pl == p_colPlayer) {
	  continue;
	}

	theHtml += "Player ";
	theHtml += ToText(pl);
	theHtml += " Strategy ";
	theHtml += ToText(profile.Get(pl)->Number());
	theHtml += " ";
      }
      theHtml += "]";
    }

    theHtml += "\n&";

    for (int st = 1; st <= NumStrats(p_colPlayer); st++) {
      theHtml += GetPlayer(p_colPlayer)->GetStrategy(st)->Name();
      if (st < NumStrats(p_colPlayer)) {
	theHtml += " & ";
      }
    } 
    theHtml += "\\\\\n";

    for (int st1 = 1; st1 <= NumStrats(p_rowPlayer); st1++) {
      profile.Set(p_rowPlayer, GetPlayer(p_rowPlayer)->GetStrategy(st1));
      theHtml += GetPlayer(p_rowPlayer)->GetStrategy(st1)->Name();
      theHtml += " & ";
      for (int st2 = 1; st2 <= NumStrats(p_colPlayer); st2++) {
	profile.Set(p_colPlayer, GetPlayer(p_colPlayer)->GetStrategy(st2));
	theHtml += " $";
	for (int pl = 1; pl <= NumPlayers(); pl++) {
	  theHtml += ToText(Payoff(GetOutcome(profile.GetIndex()+1), pl));
	  if (pl < NumPlayers()) {
	    theHtml += ",";
	  }
	}
	theHtml += "$ ";
	if (st2 < NumStrats(p_colPlayer)) {
	  theHtml += " & ";
	}
      }
      if (st1 < NumStrats(p_rowPlayer)) {
	theHtml += "\\\\\n";
      }
    } 

    theHtml += "\n\\end{game}";
  } while (iter.NextContingency());

  p_file << theHtml;
}

gbtNfgOutcome *gbtNfgGame::NewOutcome(void)
{
  gbtNfgOutcome *outcome = new gbtNfgOutcome(outcomes.Length() + 1, this);
  outcomes.Append(outcome);
  return outcome;
}

void gbtNfgGame::DeleteOutcome(gbtNfgOutcome *outcome)
{
  for (int i = 1; i <= results.Length(); i++) {
    if (results[i] == outcome)
      results[i] = 0;
  }

  delete outcomes.Remove(outcome->GetNumber());

  for (int outc = 1; outc <= outcomes.Length(); outc++)
    outcomes[outc]->number = outc;
}

const gArray<Strategy *> &gbtNfgGame::Strategies(int p) const
{
  return (players[p]->Strategies());
}

void gbtNfgGame::SetTitle(const std::string &s) 
{
  title = s; 
}

const std::string &gbtNfgGame::GetTitle(void) const 
{ return title; }

void gbtNfgGame::SetComment(const std::string &s)
{
  comment = s; 
}

const std::string &gbtNfgGame::GetComment(void) const
{ return comment; }


int gbtNfgGame::NumPlayers(void) const 
{ 
  return (players.Length()); 
}

gbtNfgPlayer *gbtNfgGame::NewPlayer(void)
{
  gbtNfgPlayer *player = new gbtNfgPlayer(players.Length() + 1, this, 1);
  players.Append(player);
  dimensions.Append(1);

  for (int outc = 1; outc <= outcomes.Length(); outc++) {
    outcomes[outc]->m_textPayoffs.Append("0");
    outcomes[outc]->m_ratPayoffs.Append(0);
    outcomes[outc]->m_doublePayoffs.Append(0.0);
  }

  return player;
}

const gArray<gbtNfgPlayer *> &gbtNfgGame::Players(void) const
{
  return players;
}

int gbtNfgGame::NumStrats(int pl) const
{
  return ((pl > 0 && pl <= NumPlayers()) ? 
	  (players[pl])->strategies.Length() : 0);
}

int gbtNfgGame::ProfileLength(void) const
{
  int nprof = 0;
  for (int i = 1; i <= players.Length(); i++)
    nprof += players[i]->strategies.Length();
  return nprof;
}

void gbtNfgGame::SetOutcome(const gArray<int> &profile, gbtNfgOutcome *outcome)
{
  int index = 1;
  for (int i = 1; i <= profile.Length(); i++)
    index += players[i]->strategies[profile[i]]->m_index;
  results[index] = outcome;
  BreakLink();
}


void gbtNfgGame::SetOutcome(const StrategyProfile &p, gbtNfgOutcome *outcome)
{
  results[p.index + 1] = outcome;
  BreakLink();
}

gbtNfgOutcome *gbtNfgGame::GetOutcome(const gArray<int> &profile) const 
{
  int index = 1;
  for (int i = 1; i <= profile.Length(); i++)
	 index += players[i]->strategies[profile[i]]->m_index;
  return results[index];
}

gbtNfgOutcome *gbtNfgGame::GetOutcome(const StrategyProfile &p) const
{
  return results[p.index + 1];
}

void gbtNfgGame::SetPayoff(gbtNfgOutcome *outcome, int pl,
			   const std::string &p_value)
{
  if (outcome) {
    outcome->m_textPayoffs[pl] = p_value;
    outcome->m_ratPayoffs[pl] = ToRational(p_value);
    outcome->m_doublePayoffs[pl] = (double) outcome->m_ratPayoffs[pl];
  }
}

gRational gbtNfgGame::Payoff(gbtNfgOutcome *outcome, int pl) const
{
  if (outcome)
    return outcome->m_ratPayoffs[pl];
  else
    return 0;
}

std::string gbtNfgGame::GetPayoffText(gbtNfgOutcome *outcome, int pl) const
{
  if (outcome) {
    return outcome->m_textPayoffs[pl];
  }
  else {
    return "0";
  }
}


// ---------------------------------------
// Nfg: Private member functions
// ---------------------------------------

void gbtNfgGame::IndexStrategies(void)
{
  long offset = 1L;

  for (int i = 1; i <= NumPlayers(); i++)  {
    int j;
    for (j = 1; j <= NumStrats(i); j++)  {
      Strategy *s = (players[i])->strategies[j];
      s->m_number = j;
      s->m_index = (j - 1) * offset;
    }
    offset *= (j - 1);
  }
}

//
// This rebuilds a new table of outcomes after the game has been
// redimensioned (change in the number of strategies).  Strategies
// numbered -1 are identified as the new strategies.
//
void gbtNfgGame::RebuildTable(void)
{
  long size = 1L;
  gArray<long> offsets(players.Length());
  for (int pl = 1; pl <= players.Length(); pl++) {
    offsets[pl] = size;
    size *= players[pl]->NumStrats();
  }

  gArray<gbtNfgOutcome *> newResults(size);
  for (int i = 1; i <= newResults.Length(); newResults[i++] = 0);

  NfgContIter iter(*this);

  do {
    long newindex = 0L;
    for (int pl = 1; pl <= players.Length(); pl++) {
      if (iter.profile[pl]->m_index < 0) {
	// This is a contingency involving a new strategy... skip
	newindex = -1L;
	break;
      }
      else {
	newindex += (iter.profile[pl]->m_number - 1) * offsets[pl];
      }
    }

    if (newindex >= 0) {
      newResults[newindex+1] = results[iter.GetIndex()+1];
    }
  } while (iter.NextContingency());

  results = newResults;

  IndexStrategies();
}

// --------------------------
// gbtNfgPlayer: Member functions 
// --------------------------

gbtNfgPlayer::gbtNfgPlayer(int n, gbtNfgGame *N, int num)
  : number(n), N(N), strategies(num)
{ 
  for (int j = 1; j <= num; j++) {
    strategies[j] = new Strategy(this);
    strategies[j]->m_number = 1;
  }
}

gbtNfgPlayer::~gbtNfgPlayer()
{ 
  for (int j = 1; j <= strategies.Length(); j++)
    delete strategies[j];
}

gbtNfgGame &gbtNfgPlayer::Game(void) const
{
  return *N;
}

const std::string &gbtNfgPlayer::GetName(void) const
{
  return name;
}

void gbtNfgPlayer::SetName(const std::string &s)
{
  name = s;
}

int gbtNfgPlayer::NumStrats(void) const 
{
  return strategies.Length();
}


const gArray<Strategy *> &gbtNfgPlayer::Strategies(void) const
{
  return strategies;
}

Strategy *gbtNfgPlayer::NewStrategy(void)
{
  Strategy *strategy = new Strategy(this);
  strategies.Append(strategy);
  strategy->m_number = strategies.Length();
  strategy->m_index = -1;   // this flags this action as new
  N->RebuildTable();
  return strategy;
}

