//
// $Source: /cvsroot/gambit/gambit/sources/game/Attic/nfg.h,v $
// $Date: 2005/06/13 02:32:10 $
// $Revision: 1.4.2.13 $
//
// DESCRIPTION:
// Declaration of normal form game representation
//
// This file is part of Gambit
// Copyright (c) 2002, The Gambit Project
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
//

#ifndef NFG_H
#define NFG_H

#include "base/base.h"
#include "math/rational.h"
#include "nfplayer.h"

class gbtNfgOutcome;
class gbtNfgPlayer;

class Strategy   {
friend class gbtNfgGame;
friend class gbtNfgPlayer;
private:
  int m_number;
  gbtNfgPlayer *m_player;
  long m_index;
  std::string m_name;

  Strategy(gbtNfgPlayer *);
  Strategy(const Strategy &);
  Strategy &operator=(const Strategy &);
  ~Strategy();

public:
  const std::string &Name(void) const { return m_name; }
  void SetName(const std::string &s)  { m_name = s; }

  gbtNfgPlayer *Player(void) const  { return m_player; }
  int Number(void) const        { return m_number; }
  long Index(void) const        { return m_index; }
};

class StrategyProfile;
class Lexicon;
class gbtEfgGame;

class gbtNfgGame  {
friend class Lexicon;
friend class NfgFileReader;
friend void SetEfg(gbtNfgGame *, const gbtEfgGame *);
friend class gbtEfgGame;
friend class gbtNfgPlayer;
protected:
  std::string title, comment;
  gArray<int> dimensions;

  gArray<gbtNfgPlayer *> players;
  gArray<gbtNfgOutcome *> outcomes;

  gArray<gbtNfgOutcome *> results;

  const gbtEfgGame *efg;

  // PRIVATE AUXILIARY MEMBER FUNCTIONS
  void IndexStrategies(void);
  int Product(const gArray<int> &);

  void BreakLink(void);

  void RebuildTable(void);
  
public:
  // CONSTRUCTORS, DESTRUCTORS, CONSTRUCTIVE OPERATORS
  gbtNfgGame(const gArray<int> &dim);
  gbtNfgGame(const gbtNfgGame &b);
  ~gbtNfgGame();
    
  // GENERAL DATA ACCESS AND MANIPULATION  
  void SetTitle(const std::string &s);
  const std::string &GetTitle(void) const;

  void SetComment(const std::string &);
  const std::string &GetComment(void) const;

  void WriteNfgFile(std::ostream &) const;
  void WriteHtmlFile(std::ostream &, int, int) const;
  void WriteOsborneFile(std::ostream &, int, int) const;

  // PLAYERS AND STRATEGIES
  int NumPlayers(void) const;
  const gArray<gbtNfgPlayer *> &Players(void) const;
  inline gbtNfgPlayer *GetPlayer(const int i) const { return Players()[i]; }
  gbtNfgPlayer *NewPlayer(void);

  int NumStrats(int pl) const;
  const gArray<int> &NumStrats(void) const  { return dimensions; }
  const gArray<Strategy *> &Strategies(int p) const;

  int ProfileLength(void) const;

  // OUTCOMES
  gbtNfgOutcome *NewOutcome(void);
  void DeleteOutcome(gbtNfgOutcome *);

  const gArray<gbtNfgOutcome *> &Outcomes(void) const  { return outcomes; }
  int NumOutcomes(void) const   { return outcomes.Length(); }

  void SetOutcome(const gArray<int> &profile, gbtNfgOutcome *outcome);
  gbtNfgOutcome *GetOutcome(const gArray<int> &profile) const;
  void SetOutcome(const StrategyProfile &p, gbtNfgOutcome *outcome);
  gbtNfgOutcome *GetOutcome(const StrategyProfile &p) const;

  void SetOutcome(int index, gbtNfgOutcome *outcome)  { results[index] = outcome; }
  gbtNfgOutcome *GetOutcome(int index) const   { return results[index]; }

  void SetPayoff(gbtNfgOutcome *, int pl, const std::string &);
  gRational Payoff(gbtNfgOutcome *, int pl) const;
  std::string GetPayoffText(gbtNfgOutcome *, int pl) const;
  gRational Payoff(gbtNfgOutcome *o, gbtNfgPlayer *p) const 
    { return Payoff(o,p->GetNumber()); }

    // defined in nfgutils.cc
  friend bool IsConstSum(const gbtNfgGame &);
  friend gRational MinPayoff(const gbtNfgGame &, int pl = 0);
  friend gRational MaxPayoff(const gbtNfgGame &, int pl = 0);

  const gbtEfgGame *AssociatedEfg(void) const   { return efg; }
};

// Exception thrown by ReadNfg if not valid .nfg file
class gbtNfgParserException : public gException {
private:
  std::string m_description;

public:
  gbtNfgParserException(void)
    : m_description("Not a valid .nfg file") { }
  gbtNfgParserException(const std::string &p_description) 
    : m_description(p_description) { }
  gbtNfgParserException(int p_line, const std::string &p_description);
  virtual ~gbtNfgParserException() { }

  std::string Description(void) const { return m_description; }
};

gbtNfgGame *ReadNfg(std::istream &);

#include "mixed.h"

class gbtNfgOutcome   {
  friend class gbtNfgGame;
  friend class MixedProfile<double>;
  friend class MixedProfile<gRational>;
  friend class MixedProfile<gNumber>;
  private:
    int number;
    gbtNfgGame *nfg;
    std::string name;
    gArray<std::string> m_textPayoffs;
    gArray<gRational> m_ratPayoffs;
    gArray<double> m_doublePayoffs;

    gbtNfgOutcome(int n, gbtNfgGame *N)
      : number(n), nfg(N),
        m_textPayoffs(nfg->NumPlayers()), 
        m_ratPayoffs(nfg->NumPlayers()),
        m_doublePayoffs(nfg->NumPlayers())
      { 
	for (int pl = 1; pl <= m_textPayoffs.Length(); pl++) {
	  m_textPayoffs[pl] = "0";
	  m_doublePayoffs[pl] = 0.0;
	}
      }
    ~gbtNfgOutcome() { }

  public:
    int GetNumber(void) const    { return number; }

    const std::string &GetName(void) const   { return name; }
    void SetName(const std::string &s)   { name = s; }

    gbtNfgGame *Game(void) const   { return nfg; }
};

#endif    // NFG_H
