//
// $Source: /cvsroot/gambit/gambit/sources/gui/gamedoc.cc,v $
// $Date: 2005/06/13 02:32:35 $
// $Revision: 1.47.2.26 $
//
// DESCRIPTION:
// Implementation of game document class
//
// This file is part of Gambit
// Copyright (c) 2005, The Gambit Project
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
//

#include <sstream>

#include <wx/wxprec.h>
#ifndef WX_PRECOMP
#include <wx/wx.h>
#endif  // WX_PRECOMP

#include "gambit.h"     // for wxGetApp()
#include "gamedoc.h"
#include "game/behav.h"

//=========================================================================
//                       class gbtBehavDominanceStack
//=========================================================================

#include "game/efstrat.h"     // for EFSupport::Undominated()

gbtBehavDominanceStack::gbtBehavDominanceStack(gbtGameDocument *p_doc,
					       bool p_strict)
  : m_doc(p_doc), m_strict(p_strict), m_noFurther(false)
{
  Reset();
}

gbtBehavDominanceStack::~gbtBehavDominanceStack()
{
  for (int i = 1; i <= m_supports.Length(); delete m_supports[i++]);
}

void gbtBehavDominanceStack::SetStrict(bool p_strict)
{
  if (m_strict != p_strict)  Reset();
  m_strict = p_strict;
}

void gbtBehavDominanceStack::Reset(void)
{
  for (int i = 1; i <= m_supports.Length(); delete m_supports[i++]);
  m_supports = gArray<gbtEfgSupport *>();
  if (m_doc->GetEfg()) {
    m_supports.Append(new gbtEfgSupport(*m_doc->GetEfg()));
    m_current = 1;
  }
}

bool gbtBehavDominanceStack::NextLevel(void)
{
  if (m_current < m_supports.Length()) {
    m_current++;
    return true;
  }

  if (m_noFurther) {
    return false;
  }

  gArray<int> players;
  for (int pl = 1; pl <= m_doc->GetEfg()->NumPlayers(); pl++) {
    players.Append(pl);
  }
  
  std::ostringstream gnull;
  gbtEfgSupport newSupport = 
    m_supports[m_current]->Undominated(m_strict, true, players, gnull);

  if (newSupport != *m_supports[m_current]) {
    m_supports.Append(new gbtEfgSupport(newSupport));
    m_current++;
    return true;
  }
  else {
    m_noFurther = true;
    return false;
  }
}

bool gbtBehavDominanceStack::PreviousLevel(void)
{
  if (m_current > 1) {
    m_current--;
    return true;
  }
  else {
    return false;
  }
}

//=========================================================================
//                   class gbtStrategyDominanceStack
//=========================================================================

#include "game/nfstrat.h"     // for NFSupport::Undominated()

gbtStrategyDominanceStack::gbtStrategyDominanceStack(gbtGameDocument *p_doc,
						     bool p_strict)
  : m_doc(p_doc), m_strict(p_strict), m_noFurther(false)
{
  Reset();
}

gbtStrategyDominanceStack::~gbtStrategyDominanceStack()
{
  for (int i = 1; i <= m_supports.Length(); delete m_supports[i++]);
}

void gbtStrategyDominanceStack::SetStrict(bool p_strict)
{
  if (m_strict != p_strict)  Reset();
  m_strict = p_strict;
}

void gbtStrategyDominanceStack::Reset(void)
{
  for (int i = 1; i <= m_supports.Length(); delete m_supports[i++]);
  m_supports = gArray<gbtNfgSupport *>();
  if (m_doc->GetNfg()) {
    m_supports.Append(new gbtNfgSupport(*m_doc->GetNfg()));
    m_current = 1;
  }
  m_noFurther = false;
}

bool gbtStrategyDominanceStack::NextLevel(void)
{
  if (m_current < m_supports.Length()) {
    m_current++;
    return true;
  }

  if (m_noFurther) {
    return false;
  }

  gArray<int> players;
  for (int pl = 1; pl <= m_doc->GetNfg()->NumPlayers(); pl++) {
    players.Append(pl);
  }
  
  std::ostringstream gnull;
  gbtNfgSupport newSupport = 
    m_supports[m_current]->Undominated(m_strict, players, gnull);

  if (newSupport != *m_supports[m_current]) {
    m_supports.Append(new gbtNfgSupport(newSupport));
    m_current++;
    return true;
  }
  else {
    m_noFurther = true;
    return false;
  }
}

bool gbtStrategyDominanceStack::PreviousLevel(void)
{
  if (m_current > 1) {
    m_current--;
    return true;
  }
  else {
    return false;
  }
}

//=========================================================================
//                          class gbtGameDocument
//=========================================================================

gbtGameDocument::gbtGameDocument(gbtEfgGame *p_efg) 
  : m_efg(p_efg), m_nfg(0),
    m_selectNode(0), m_modified(false),
    m_behavSupports(this, true), m_stratSupports(this, true),
    m_currentProfile(0)
{
  wxGetApp().AddDocument(this);

  std::ostringstream s;
  m_efg->WriteEfgFile(s);
  m_undoList.Append(s.str());
}

gbtGameDocument::gbtGameDocument(gbtNfgGame *p_nfg) 
  : m_efg(0), m_nfg(p_nfg),
    m_selectNode(0), m_modified(false),
    m_behavSupports(this, true), m_stratSupports(this, true),
    m_currentProfile(0)
{
  wxGetApp().AddDocument(this);

  std::ostringstream s;
  m_nfg->WriteNfgFile(s);
  m_undoList.Append(s.str());
}

gbtGameDocument::~gbtGameDocument()
{
  wxGetApp().RemoveDocument(this);
}

void gbtGameDocument::UpdateViews(bool p_setModified)
{
  if (p_setModified)  {
    m_modified = true;
    m_behavSupports.Reset();
    m_stratSupports.Reset();
    m_behavProfiles = gList<BehavProfile<gNumber> >();
    m_mixedProfiles = gList<MixedProfile<gNumber> >();
    m_currentProfile = 0;
    m_redoList = gList<std::string>();

    std::ostringstream s;
    if (m_efg) {
      m_efg->WriteEfgFile(s);
    }
    else {
      m_nfg->WriteNfgFile(s);
    }
    m_undoList.Append(s.str());
  }

  for (int i = 1; i <= m_views.Length(); m_views[i++]->OnUpdate());
}

void gbtGameDocument::BuildNfg(void)
{ 
  if (m_efg && m_efg->AssociatedNfg() == 0) {
    MakeReducedNfg(gbtEfgSupport(*m_efg));
    m_stratSupports.Reset();
    for (int i = 1; i <= m_behavProfiles.Length(); i++) {
      m_mixedProfiles.Append(MixedProfile<gNumber>(m_behavProfiles[i]));
    }
  }
}

//
// A word about the undo and redo features:
// We store a list of the textual representation of games.  We don't
// store other aspects of the state (e.g., profiles) as yet.
// The "undo" list includes the representation of the current state
// of the game (hence, CanUndo() only returns true when the list has
// more than one element.
//
void gbtGameDocument::Undo(void)
{
  // The current game is at the end of the undo list; move it to the redo list
  m_redoList.Append(m_undoList[m_undoList.Length()]);
  m_undoList.Remove(m_undoList.Length());

  if (m_efg) {
    delete m_efg;
    std::istringstream s(m_undoList[m_undoList.Length()]);
    m_efg = ReadEfgFile(s);
  }
  else {
    delete m_nfg;
    m_nfg = 0;
    std::istringstream s(m_undoList[m_undoList.Length()]);
    m_nfg = ReadNfg(s);
  }

  m_behavSupports.Reset();
  m_stratSupports.Reset();
  m_behavProfiles = gList<BehavProfile<gNumber> >();
  m_mixedProfiles = gList<MixedProfile<gNumber> >();
  m_currentProfile = 0;

  for (int i = 1; i <= m_views.Length(); m_views[i++]->OnUpdate());
}

void gbtGameDocument::Redo(void)
{
  m_undoList.Append(m_redoList[m_redoList.Length()]);
  m_redoList.Remove(m_redoList.Length());

  if (m_efg) {
    delete m_efg;
    std::istringstream s(m_undoList[m_undoList.Length()]);
    m_efg = ReadEfgFile(s);
  }
  else {
    delete m_nfg;
    m_nfg = 0;
    std::istringstream s(m_undoList[m_undoList.Length()]);
    m_nfg = ReadNfg(s);
  }

  m_behavSupports.Reset();
  m_stratSupports.Reset();
  m_behavProfiles = gList<BehavProfile<gNumber> >();
  m_mixedProfiles = gList<MixedProfile<gNumber> >();
  m_currentProfile = 0;

  for (int i = 1; i <= m_views.Length(); m_views[i++]->OnUpdate());
}


void gbtGameDocument::SetCurrentProfile(int p_profile)
{
  m_currentProfile = p_profile;
  UpdateViews(false);
}

void gbtGameDocument::AddProfiles(const gList<BehavProfile<gNumber> > &p_profiles)
{
  for (int i = 1; i <= p_profiles.Length(); i++) {
    m_behavProfiles.Append(p_profiles[i]);
    if (GetNfg() != 0) {
      m_mixedProfiles.Append(MixedProfile<gNumber>(p_profiles[i]));
    }
  }

  m_currentProfile = m_behavProfiles.Length();
  UpdateViews(false);
}

void gbtGameDocument::AddProfile(const BehavProfile<gNumber> &p_profile)
{
  m_behavProfiles.Append(p_profile);
  if (GetNfg() != 0) {
    m_mixedProfiles.Append(MixedProfile<gNumber>(p_profile));
  }
  
  m_currentProfile = m_behavProfiles.Length();
  UpdateViews(false);
}

void gbtGameDocument::AddProfiles(const gList<MixedProfile<gNumber> > &p_profiles)
{
  for (int i = 1; i <= p_profiles.Length(); i++) {
    m_mixedProfiles.Append(p_profiles[i]);
    if (m_efg != 0) {
      m_behavProfiles.Append(BehavProfile<gNumber>(p_profiles[i]));
    }
  }

  m_currentProfile = m_mixedProfiles.Length();
  UpdateViews(false);
}

void gbtGameDocument::AddProfile(const MixedProfile<gNumber> &p_profile)
{
  m_mixedProfiles.Append(p_profile);
  if (m_efg != 0) {
    m_behavProfiles.Append(BehavProfile<gNumber>(p_profile));
  }

  m_currentProfile = m_mixedProfiles.Length();
  UpdateViews(false);
}

void gbtGameDocument::SetBehavElimStrength(bool p_strict)
{
  m_behavSupports.SetStrict(p_strict);
  UpdateViews(false);
}

bool gbtGameDocument::NextBehavElimLevel(void)
{
  bool ret = m_behavSupports.NextLevel();
  UpdateViews(false);
  return ret;
}

void gbtGameDocument::PreviousBehavElimLevel(void)
{
  m_behavSupports.PreviousLevel();
  UpdateViews(false);
}

void gbtGameDocument::TopBehavElimLevel(void)
{
  m_behavSupports.TopLevel();
  UpdateViews(false);
}

bool gbtGameDocument::CanBehavElim(void) const
{
  return m_behavSupports.CanEliminate();
}

int gbtGameDocument::GetBehavElimLevel(void) const
{
  return m_behavSupports.GetLevel();
}


void gbtGameDocument::SetStrategyElimStrength(bool p_strict)
{
  m_stratSupports.SetStrict(p_strict);
  UpdateViews(false);
}

bool gbtGameDocument::NextStrategyElimLevel(void)
{
  bool ret = m_stratSupports.NextLevel();
  UpdateViews(false);
  return ret;
}

void gbtGameDocument::PreviousStrategyElimLevel(void)
{
  m_stratSupports.PreviousLevel();
  UpdateViews(false);
}

void gbtGameDocument::TopStrategyElimLevel(void)
{
  m_stratSupports.TopLevel();
  UpdateViews(false);
}

bool gbtGameDocument::CanStrategyElim(void) const
{
  return m_stratSupports.CanEliminate();
}

int gbtGameDocument::GetStrategyElimLevel(void) const
{
  return m_stratSupports.GetLevel();
}


void gbtGameDocument::SetSelectNode(gbtEfgNode *p_node)
{
  m_selectNode = p_node;
  UpdateViews(false);
}

std::string gbtGameDocument::GetRealizProb(const gbtEfgNode *p_node) const
{
  if (GetCurrentProfile() == 0 || !p_node) {
    return "";
  }
  return ToText(GetBehavProfiles()[GetCurrentProfile()].GetRealizProb(p_node),
		m_style.NumDecimals());
}

std::string gbtGameDocument::GetBeliefProb(const gbtEfgNode *p_node) const
{
  if (GetCurrentProfile() == 0 || !p_node || !p_node->GetPlayer()) {
    return "";
  }

  if (GetBehavProfiles()[GetCurrentProfile()].GetIsetProb(p_node->GetInfoset()) > gNumber(0)) {
    return ToText(GetBehavProfiles()[GetCurrentProfile()].GetBeliefProb(p_node),
		  m_style.NumDecimals());
  }
  else {
    // We don't compute assessments yet!
    return "*";
  }
}

std::string gbtGameDocument::GetNodeValue(const gbtEfgNode *p_node,
					  int p_player) const
{
  if (GetCurrentProfile() == 0 || !p_node) {
    return "";
  }

  return ToText(GetBehavProfiles()[GetCurrentProfile()].GetNodeValue(p_node)[p_player], 
		m_style.NumDecimals());
}

std::string gbtGameDocument::GetInfosetProb(const gbtEfgNode *p_node) const
{
  if (GetCurrentProfile() == 0 || !p_node || !p_node->GetPlayer()) {
    return "";
  }
  return ToText(GetBehavProfiles()[GetCurrentProfile()].GetIsetProb(p_node->GetInfoset()),
		m_style.NumDecimals());
}

std::string gbtGameDocument::GetInfosetValue(const gbtEfgNode *p_node) const
{
  if (GetCurrentProfile() == 0 || !p_node || !p_node->GetPlayer() ||
      p_node->GetPlayer()->IsChance()) {
    return "";
  }
  if (GetBehavProfiles()[GetCurrentProfile()].GetIsetProb(p_node->GetInfoset()) > gNumber(0)) {
    return ToText(GetBehavProfiles()[GetCurrentProfile()].GetIsetValue(p_node->GetInfoset()),
		  m_style.NumDecimals());
  }
  else {
    // In the absence of beliefs, this is not well-defined in general
    return "*";
  }
}

std::string gbtGameDocument::GetActionProb(const gbtEfgNode *p_node, int p_act) const
{
  if (p_node->GetPlayer() && p_node->GetPlayer()->IsChance()) {
    return GetEfg()->GetChanceProbText(p_node->GetInfoset(), p_act);
  }

  if (GetCurrentProfile() == 0 || !p_node->GetPlayer()) {
    return "";
  }
  
  const BehavProfile<gNumber> &profile = GetBehavProfiles()[GetCurrentProfile()];

  if (!profile.IsDefinedAt(p_node->GetInfoset())) {
    return "*";
  }
  
  return ToText(GetBehavProfiles()[GetCurrentProfile()].GetActionProb(p_node->GetInfoset()->Actions()[p_act]),
		m_style.NumDecimals());
}

std::string gbtGameDocument::GetActionValue(const gbtEfgNode *p_node, int p_act) const
{
  if (GetCurrentProfile() == 0 || !p_node || !p_node->GetPlayer() ||
      p_node->GetPlayer()->IsChance()) {
    return "";
  }

  if (GetBehavProfiles()[GetCurrentProfile()].GetIsetProb(p_node->GetInfoset()) > gNumber(0)) {
    return ToText(GetBehavProfiles()[GetCurrentProfile()].GetActionValue(p_node->GetInfoset()->Actions()[p_act]),
		  m_style.NumDecimals());
  }
  else  {
    // In the absence of beliefs, this is not well-defined
    return "*";
  }
}

gNumber gbtGameDocument::ActionProb(const gbtEfgNode *p_node, int p_action) const
{
  if (p_node->GetPlayer() && p_node->GetPlayer()->IsChance()) {
    return GetEfg()->GetChanceProb(p_node->GetInfoset(), p_action);
  }

  if (GetCurrentProfile() && p_node->GetInfoset()) {
    return GetBehavProfiles()[GetCurrentProfile()](p_node->GetInfoset()->GetPlayer()->GetNumber(), p_node->GetInfoset()->GetNumber(), p_action);
  }
  return -1;
}

