//
// $Source: /cvsroot/gambit/gambit/sources/gui/Attic/nfgpanel.cc,v $
// $Date: 2005/06/13 16:49:15 $
// $Revision: 1.1.2.23 $
//
// DESCRIPTION:
// Implementation of normal form panel
//
// This file is part of Gambit
// Copyright (c) 2005, The Gambit Project
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
//

#include <wx/wxprec.h>
#ifndef WX_PRECOMP
#include <wx/wx.h>
#endif  // WX_PRECOMP
#include <wx/dnd.h>     // for drag-and-drop features
#include <wx/image.h>

#include "nfgpanel.h"
#include "menuconst.h"

#include "sheet.h"
#include "renratio.h"    // special renderer for rational numbers


#include "../bitmaps/person.xpm"

class gbtTablePlayerIcon : public wxStaticBitmap {
private:
  int m_player;

  // Event handlers
  void OnLeftClick(wxMouseEvent &);

public:
  gbtTablePlayerIcon(wxWindow *p_parent, int p_player);

  DECLARE_EVENT_TABLE()
};

BEGIN_EVENT_TABLE(gbtTablePlayerIcon, wxStaticBitmap)
  EVT_LEFT_DOWN(gbtTablePlayerIcon::OnLeftClick)
END_EVENT_TABLE()

gbtTablePlayerIcon::gbtTablePlayerIcon(wxWindow *p_parent, int p_player)
  : wxStaticBitmap(p_parent, -1, wxBitmap(person_xpm)), m_player(p_player)
{ }

void gbtTablePlayerIcon::OnLeftClick(wxMouseEvent &)
{
  wxBitmap bitmap(person_xpm);

#ifdef __WXMSW__
  wxImage image = bitmap.ConvertToImage();
#else
  wxIcon image;
  image.CopyFromBitmap(bitmap);
#endif // _WXMSW__
      
  wxTextDataObject textData(wxString::Format(wxT("P%d"), m_player));
  wxDropSource source(textData, this, image, image, image);
  source.DoDragDrop(wxDrag_DefaultMove);
}

class gbtTablePlayerPanel : public wxPanel {
private:
  gbtNfgPanel *m_nfgPanel;
  gbtGameDocument *m_doc;
  int m_player;
  wxStaticText *m_playerLabel, *m_payoff;
  wxChoice *m_strategies;

  // Event handlers
  void OnStrategySelect(wxCommandEvent &);
  
public:
  gbtTablePlayerPanel(wxWindow *, gbtNfgPanel *, 
		      gbtGameDocument *, int p_player);

  void OnUpdate(void);
};

gbtTablePlayerPanel::gbtTablePlayerPanel(wxWindow *p_parent,
					 gbtNfgPanel *p_nfgPanel,
					 gbtGameDocument *p_doc,
					 int p_player)
  : wxPanel(p_parent, -1), 
    m_nfgPanel(p_nfgPanel), m_doc(p_doc), m_player(p_player)
{
  SetBackgroundColour(wxColour(215, 215, 215));

  wxBoxSizer *topSizer = new wxBoxSizer(wxVERTICAL);

  wxBoxSizer *labelSizer = new wxBoxSizer(wxHORIZONTAL);

  wxStaticBitmap *playerIcon = new gbtTablePlayerIcon(this, m_player);
  labelSizer->Add(playerIcon, 0, wxALL | wxALIGN_CENTER, 0);
  m_playerLabel = new wxStaticText(this, wxID_STATIC, wxT(""),
				   wxDefaultPosition, wxSize(125, -1),
				   wxALIGN_LEFT);
  m_playerLabel->SetFont(wxFont(10, wxSWISS, wxNORMAL, wxBOLD));
  labelSizer->Add(m_playerLabel, 1, wxLEFT | wxALIGN_CENTER, 10);
  topSizer->Add(labelSizer, 0, wxALL, 0);

  m_strategies = new wxChoice(this, -1);
  Connect(m_strategies->GetId(), wxEVT_COMMAND_CHOICE_SELECTED,
	  wxCommandEventHandler(gbtTablePlayerPanel::OnStrategySelect));
  topSizer->Add(m_strategies, 0, wxALL | wxEXPAND, 0);

  m_payoff = new wxStaticText(this, wxID_STATIC, wxT("Payoff:"),
			      wxDefaultPosition, wxDefaultSize,
			      wxALIGN_CENTER | wxST_NO_AUTORESIZE);
  m_payoff->SetFont(wxFont(10, wxSWISS, wxNORMAL, wxNORMAL));
  topSizer->Add(m_payoff, 0, wxALL | wxEXPAND, 0);
  topSizer->Show(m_payoff, false);

  SetSizer(topSizer);
  topSizer->SetSizeHints(this);
  topSizer->Fit(this);
  Layout();
}

void gbtTablePlayerPanel::OnUpdate(void)
{
  if (!m_doc->GetNfg())  return;

  wxColour color = m_doc->GetStyle().GetPlayerColor(m_player);

  m_playerLabel->SetForegroundColour(color);
  m_playerLabel->SetLabel(wxString(m_doc->GetNfg()->GetPlayer(m_player)->GetName().c_str(),
				   *wxConvCurrent));

  const gbtNfgSupport &support = m_doc->GetNfgSupport();
  m_strategies->Clear();
  m_strategies->SetForegroundColour(color);

  for (int st = 1; st <= support.NumStrats(m_player); st++) {
    Strategy *strategy = support.GetStrategy(m_player, st);
    std::string label = strategy->Name();
    std::string pay = "";

    if (m_doc->GetCurrentProfile() > 0) {
      const MixedProfile<gNumber> &profile =
	m_doc->GetMixedProfiles()[m_doc->GetCurrentProfile()];

      pay = " (" + ToText(profile.Payoff(m_player, strategy),
			 m_doc->GetStyle().NumDecimals()) + ")";
    }

    m_strategies->Append(wxString((label + pay).c_str(), *wxConvCurrent));
  }

  m_strategies->SetSelection(m_nfgPanel->GetContingency()[m_player]-1);

  if (m_doc->GetCurrentProfile() > 0) {
    m_payoff->SetForegroundColour(color);

    const MixedProfile<gNumber> &profile =
      m_doc->GetMixedProfiles()[m_doc->GetCurrentProfile()];

    std::string pay = ToText(profile.Payoff(m_player),
			     m_doc->GetStyle().NumDecimals());
    m_payoff->SetLabel(wxT("Payoff: ") + 
		       wxString(pay.c_str(), *wxConvCurrent));
    GetSizer()->Show(m_payoff, true);
  }
  else {
    GetSizer()->Show(m_payoff, false);
  }

  GetSizer()->Layout();
  GetSizer()->SetSizeHints(this);
  GetSizer()->Fit(this);
}

void gbtTablePlayerPanel::OnStrategySelect(wxCommandEvent &p_event)
{
  m_nfgPanel->SetStrategy(m_player, p_event.GetSelection() + 1);
}

//=====================================================================
//                  class gbtTablePlayerToolbar
//=====================================================================

class gbtTablePlayerToolbar : public wxPanel, public gbtGameView {
private:
  gbtNfgPanel *m_nfgPanel;
  gArray<gbtTablePlayerPanel *> m_playerPanels;

  // Implementation of gbtGameView members
  void OnUpdate(void);

public:
  gbtTablePlayerToolbar(gbtNfgPanel *p_parent, gbtGameDocument *p_doc);
};


gbtTablePlayerToolbar::gbtTablePlayerToolbar(gbtNfgPanel *p_parent, 
					     gbtGameDocument *p_doc)
  : wxPanel(p_parent, -1, wxDefaultPosition, wxSize(110, -1)), 
    gbtGameView(p_doc),
    m_nfgPanel(p_parent)
{ 
  SetBackgroundColour(wxColour(215, 215, 215));

  wxBoxSizer *topSizer = new wxBoxSizer(wxVERTICAL);

  for (int pl = 1; pl <= m_doc->NumPlayers(); pl++) {
    m_playerPanels.Append(new gbtTablePlayerPanel(this, p_parent, m_doc, pl));
    topSizer->Add(m_playerPanels[pl], 0, wxALL | wxEXPAND, 5);
  }

  SetSizer(topSizer);
  Layout();
}

void gbtTablePlayerToolbar::OnUpdate(void)
{
  if (!m_doc->GetNfg())  return;

  while (m_playerPanels.Length() < m_doc->NumPlayers()) {
    gbtTablePlayerPanel *panel = 
      new gbtTablePlayerPanel(this, m_nfgPanel, m_doc,
			      m_playerPanels.Length()+1);
    m_playerPanels.Append(panel);
    GetSizer()->Add(panel, 0, wxALL | wxEXPAND, 5);
  }

  while (m_playerPanels.Length() > m_doc->NumPlayers()) {
    gbtTablePlayerPanel *panel = m_playerPanels.Remove(m_playerPanels.Length());
    GetSizer()->Remove(panel);
    panel->Destroy();
  }
  
  for (int pl = 1; pl <= m_playerPanels.Length(); pl++) {
    m_playerPanels[pl]->OnUpdate();
  }

  GetSizer()->Layout();
}


//--------------------------------------------------------------------------
//                 class gbtPlayerLabelDropTarget
//--------------------------------------------------------------------------

class gbtPlayerLabelDropTarget : public wxTextDropTarget {
private:
  gbtPlayerLabelCtrl *m_owner;

public:
  gbtPlayerLabelDropTarget(gbtPlayerLabelCtrl *p_owner) 
    : m_owner(p_owner) { }

  bool OnDropText(wxCoord x, wxCoord y, const wxString &p_text);
};

bool gbtPlayerLabelDropTarget::OnDropText(wxCoord p_x, wxCoord p_y,
					  const wxString &p_text)
{
  if (p_text[0] == 'P') {
    long pl;
    p_text.Right(p_text.Length() - 1).ToLong(&pl);
    
    if (pl == m_owner->GetPlayer()) {
      return false;
    }

    if (m_owner->IsRowPlayer()) {
      wxCommandEvent event(GBT_ROW_PLAYER_CHANGE);
      event.SetInt(pl);
      wxPostEvent(m_owner, event);
    }
    else {
      wxCommandEvent event(GBT_COL_PLAYER_CHANGE);
      event.SetInt(pl);
      wxPostEvent(m_owner, event);
    }
    return true;
  }
  return false;
}

//-----------------------------------------------------------------------
//                  class gbtPlayerLabelCtrl
//-----------------------------------------------------------------------

BEGIN_EVENT_TABLE(gbtPlayerLabelCtrl, wxScrolledWindow)
  EVT_RIGHT_DOWN(gbtPlayerLabelCtrl::OnRightDown)
  EVT_MENU_RANGE(1000, 32767, gbtPlayerLabelCtrl::OnMenu)
  EVT_ENTER_WINDOW(gbtPlayerLabelCtrl::OnEnterWindow)
  EVT_LEAVE_WINDOW(gbtPlayerLabelCtrl::OnLeaveWindow)
END_EVENT_TABLE()

gbtPlayerLabelCtrl::gbtPlayerLabelCtrl(wxWindow *p_parent,
				       gbtGameDocument *p_doc,
				       bool p_isRowPlayer)
  : wxScrolledWindow(p_parent, -1), gbtGameView(p_doc),
    m_isRowPlayer(p_isRowPlayer), m_player(1),
    m_popupMenu(0)
{
  SetBackgroundColour(wxColour(250, 250, 250));
  if (p_isRowPlayer) {
    SetToolTip(_("The player choosing the row"));
  }
  else {
    SetToolTip(_("The player choosing the column"));
  }
  SetDropTarget(new gbtPlayerLabelDropTarget(this));
}

gbtPlayerLabelCtrl::~gbtPlayerLabelCtrl()
{
  if (m_popupMenu)  delete m_popupMenu;
}

void gbtPlayerLabelCtrl::SetPlayer(int p_player)
{
  m_player = p_player;
  Refresh();
}

void gbtPlayerLabelCtrl::OnDraw(wxDC &p_dc)
{
  if (!m_doc->GetNfg())  m_doc->BuildNfg();

  p_dc.SetFont(wxFont(12, wxSWISS, wxNORMAL, wxBOLD));
  wxCoord width, height;

  wxString label = wxString(m_doc->GetNfg()->GetPlayer(m_player)->GetName().c_str(),
			    *wxConvCurrent);
  wxColour color = m_doc->GetStyle().GetPlayerColor(m_player);
  
  p_dc.GetTextExtent(label, &width, &height);
  p_dc.SetTextForeground(color);
  wxSize size = GetClientSize();

  if (m_isRowPlayer) {
    // This is to fix a strange crash under GTK with blank labels
    if (height == 0 || width == 0)  return;
    p_dc.DrawRotatedText(label, 
			 (size.GetWidth() - height) / 2,
			 (size.GetHeight() + width) / 2,
			 90);
  }
  else {
    p_dc.DrawText(label, 
		  (size.GetWidth() - width) / 2,
		  (size.GetHeight() - height) / 2);
  }
}

void gbtPlayerLabelCtrl::OnRightDown(wxMouseEvent &p_event)
{
  if (m_popupMenu)  delete m_popupMenu;

  m_popupMenu = new wxMenu(_T("Show player on this axis"));
  gbtNfgGame *game = m_doc->GetNfg();
  for (int pl = 1; pl <= game->NumPlayers(); pl++) {
    m_popupMenu->Append(1000 + pl,
			wxString::Format(wxT("%d: "), pl) +
			wxString(game->GetPlayer(pl)->GetName().c_str(),
				 *wxConvCurrent));
  }

  PopupMenu(m_popupMenu, p_event.GetX(), p_event.GetY());
}

void gbtPlayerLabelCtrl::OnMenu(wxCommandEvent &p_event)
{
  int newPlayer = p_event.GetId() - 1000;

  if (m_isRowPlayer) {
    wxCommandEvent event(GBT_ROW_PLAYER_CHANGE);
    event.SetInt(newPlayer);
    wxPostEvent(this, event);
  }
  else {
    wxCommandEvent event(GBT_COL_PLAYER_CHANGE);
    event.SetInt(newPlayer);
    wxPostEvent(this, event);
  }
}

void gbtPlayerLabelCtrl::OnEnterWindow(wxMouseEvent &)
{
  SetBackgroundColour(wxColour(210, 210, 210));
  Refresh();
}

void gbtPlayerLabelCtrl::OnLeaveWindow(wxMouseEvent &)
{
  SetBackgroundColour(wxColour(250, 250, 250));
  Refresh();
}

//=====================================================================
//               class gbtStrategyDominanceToolbar
//=====================================================================

class gbtStrategyDominanceToolbar : public wxPanel, public gbtGameView {
private:
  wxButton *m_topButton, *m_prevButton, *m_nextButton, *m_allButton;
  wxStaticText *m_level;

  // Overriding gbtGameView members
  void OnUpdate(void);

  // Event handlers
  void OnStrength(wxCommandEvent &);
  void OnTopLevel(wxCommandEvent &);
  void OnPreviousLevel(wxCommandEvent &);
  void OnNextLevel(wxCommandEvent &);
  void OnLastLevel(wxCommandEvent &);

public:
  gbtStrategyDominanceToolbar(wxWindow *p_parent, gbtGameDocument *p_doc);
  virtual ~gbtStrategyDominanceToolbar() { }
};


#include "../bitmaps/next.xpm"
#include "../bitmaps/prev.xpm"
#include "../bitmaps/tobegin.xpm"
#include "../bitmaps/toend.xpm"

gbtStrategyDominanceToolbar::gbtStrategyDominanceToolbar(wxWindow *p_parent,
							 gbtGameDocument *p_doc)
  : wxPanel(p_parent, -1), gbtGameView(p_doc)
{
  wxBoxSizer *topSizer = new wxBoxSizer(wxHORIZONTAL);

  topSizer->Add(new wxStaticText(this, wxID_STATIC,
				 wxT("Hide strategies which are ")),
		0, wxALL | wxALIGN_CENTER, 5);

  wxString domChoices[] = { wxT("strictly"), wxT("strictly or weakly") };
  wxChoice *choice = new wxChoice(this, -1, wxDefaultPosition, wxDefaultSize,
				  2, domChoices);
  choice->SetSelection(0);
  Connect(choice->GetId(), wxEVT_COMMAND_CHOICE_SELECTED,
	  wxCommandEventHandler(gbtStrategyDominanceToolbar::OnStrength));
  topSizer->Add(choice, 0, wxALL | wxALIGN_CENTER, 5);

  topSizer->Add(new wxStaticText(this, wxID_STATIC, wxT("dominated:")),
		0, wxALL | wxALIGN_CENTER, 5);

  m_topButton = new wxBitmapButton(this, -1, wxBitmap(tobegin_xpm));
  m_topButton->SetToolTip(_("Show all strategies"));
  Connect(m_topButton->GetId(), wxEVT_COMMAND_BUTTON_CLICKED,
	  wxCommandEventHandler(gbtStrategyDominanceToolbar::OnTopLevel));
  topSizer->Add(m_topButton, 0, wxALL | wxALIGN_CENTER, 5);

  m_prevButton = new wxBitmapButton(this, -1, wxBitmap(prev_xpm));
  m_prevButton->SetToolTip(_("Previous round of elimination"));
  Connect(m_prevButton->GetId(), wxEVT_COMMAND_BUTTON_CLICKED,
	  wxCommandEventHandler(gbtStrategyDominanceToolbar::OnPreviousLevel));
  topSizer->Add(m_prevButton, 0, wxALL | wxALIGN_CENTER, 5);

  m_level = new wxStaticText(this, wxID_STATIC, 
			     wxT("All strategies shown"),
			     wxDefaultPosition, wxDefaultSize,
			     wxALIGN_CENTER | wxST_NO_AUTORESIZE);
  topSizer->Add(m_level, 0, wxALL | wxALIGN_CENTER, 5);

  m_nextButton = new wxBitmapButton(this, -1, wxBitmap(next_xpm));
  m_nextButton->SetToolTip(_("Next round of elimination"));
  Connect(m_nextButton->GetId(), wxEVT_COMMAND_BUTTON_CLICKED,
	  wxCommandEventHandler(gbtStrategyDominanceToolbar::OnNextLevel));
  topSizer->Add(m_nextButton, 0, wxALL | wxALIGN_CENTER, 5);

  m_allButton = new wxBitmapButton(this, -1, wxBitmap(toend_xpm));
  m_allButton->SetToolTip(_("Eliminate iteratively"));
  Connect(m_allButton->GetId(), wxEVT_COMMAND_BUTTON_CLICKED,
	  wxCommandEventHandler(gbtStrategyDominanceToolbar::OnLastLevel));
  topSizer->Add(m_allButton, 0, wxALL | wxALIGN_CENTER, 5);

  SetSizer(topSizer);
  Layout();
}

void gbtStrategyDominanceToolbar::OnStrength(wxCommandEvent &p_event)
{
  m_doc->SetStrategyElimStrength(p_event.GetSelection() == 0);
}

void gbtStrategyDominanceToolbar::OnTopLevel(wxCommandEvent &)
{
  m_doc->TopStrategyElimLevel();
}

void gbtStrategyDominanceToolbar::OnPreviousLevel(wxCommandEvent &)
{
  m_doc->PreviousStrategyElimLevel();
}

void gbtStrategyDominanceToolbar::OnNextLevel(wxCommandEvent &)
{
  m_doc->NextStrategyElimLevel();
}

void gbtStrategyDominanceToolbar::OnLastLevel(wxCommandEvent &)
{
  while (m_doc->NextStrategyElimLevel());
}

void gbtStrategyDominanceToolbar::OnUpdate(void)
{
  m_topButton->Enable(m_doc->GetStrategyElimLevel() > 1);
  m_prevButton->Enable(m_doc->GetStrategyElimLevel() > 1);
  m_nextButton->Enable(m_doc->CanStrategyElim());
  m_allButton->Enable(m_doc->CanStrategyElim());
  if (m_doc->GetStrategyElimLevel() == 1) {
    m_level->SetLabel(wxT("All strategies shown"));
  }
  else if (m_doc->GetStrategyElimLevel() == 2) {
    m_level->SetLabel(wxT("Eliminated 1 level")); 
  }
  else {
    m_level->SetLabel(wxString::Format(wxT("Eliminated %d levels"), 
				       m_doc->GetStrategyElimLevel()-1));
  }
  GetSizer()->Layout();
}

//======================================================================
//                        class gbtMatrixSheet
//======================================================================

class gbtMatrixSheet : public wxSheet {
private:
  gbtGameDocument *m_doc;
  gbtNfgPanel *m_nfgPanel;

  // Overriding wxSheet members for data access
  wxString GetRowLabelValue(int);
  wxString GetColLabelValue(int);
  wxString GetCellValue(const wxSheetCoords &);

  void SetCellValue(const wxSheetCoords &, const wxString &);
  wxSheetCellAttr GetAttr(const wxSheetCoords &p_coords, wxSheetAttr_Type) const;

  // Overriding wxSheet members to disable selection behavior
  bool SelectRow(int, bool = false, bool = false)
    { return false; }
  bool SelectRows(int, int, bool = false, bool = false)
    { return false; }
  bool SelectCol(int, bool = false, bool = false)
    { return false; }
  bool SelectCols(int, int, bool = false, bool = false)
    { return false; }
  bool SelectCell(const wxSheetCoords&, bool = false, bool = false)
    { return false; }
  bool SelectBlock(const wxSheetBlock&, bool = false, bool = false)
    { return false; }
  bool SelectAll(bool = false) { return false; }


  bool HasSelection(bool = true) const { return false; }
  bool IsCellSelected(const wxSheetCoords &) const { return false; }
  bool IsRowSelected(int) const { return false; }
  bool IsColSelected(int) const { return false; }
  bool DeselectBlock(const wxSheetBlock &, bool = false) { return false; }
  bool ClearSelection(bool = false) { return false; }

  // Overriding wxSheet member to suppress drawing of cursor
  void DrawCursorCellHighlight(wxDC&, const wxSheetCellAttr &) { }

  // Overriding wxSheet members for custom drawing
  void DrawCellBorder(wxDC& dc, const wxSheetCoords& coords);
  wxRect CellToRect(const wxSheetCoords &p_coords,
		    bool p_getDeviceRect = false) const;
  void DrawColLabels(wxDC& dc, const wxArrayInt& cols);

  // Event handlers
  void OnColSize(wxSheetEvent &);
  void OnRowSize(wxSheetEvent &);

public:
  gbtMatrixSheet(gbtNfgPanel *p_parent, wxWindowID p_id,
		 gbtGameDocument *p_doc);

  void OnUpdate(void);
};

gbtMatrixSheet::gbtMatrixSheet(gbtNfgPanel *p_parent, wxWindowID p_id,
			       gbtGameDocument *p_doc)
  : wxSheet(p_parent, p_id), m_doc(p_doc),
    m_nfgPanel(p_parent)
{
  if (m_doc->GetNfg()) {
    CreateGrid(m_doc->GetNfg()->GetPlayer(m_nfgPanel->GetRowPlayer())->NumStrats(),
	       m_doc->GetNfg()->GetPlayer(m_nfgPanel->GetColPlayer())->NumStrats()*
	       m_doc->NumPlayers());
  }
  else {
    CreateGrid(0, 0);
  }

  EnableEditing(!m_doc->GetEfg());
  SetCursorCellHighlightColour(*wxWHITE);

  Connect(GetId(), wxEVT_SHEET_COL_SIZE,
	  (wxObjectEventFunction) (wxEventFunction) wxStaticCastEvent(wxSheetEventFunction, wxSheetEventFunction(&gbtMatrixSheet::OnColSize)));

  Connect(GetId(), wxEVT_SHEET_ROW_SIZE,
	  (wxObjectEventFunction) (wxEventFunction) wxStaticCastEvent(wxSheetEventFunction, wxSheetEventFunction(&gbtMatrixSheet::OnRowSize)));
}

void gbtMatrixSheet::OnColSize(wxSheetEvent &p_event)
{
  // Keep all columns at the same width
  int width = GetColWidth(p_event.GetCol());
  SetDefaultColWidth(width, true);
  Refresh();
}

void gbtMatrixSheet::OnRowSize(wxSheetEvent &p_event)
{
  // Keep all rows at the same width
  int height = GetRowHeight(p_event.GetRow());
  SetDefaultRowHeight(height, true);
  Refresh();
}

wxString gbtMatrixSheet::GetRowLabelValue(int p_row)
{
  int numStrats = m_doc->GetNfgSupport().NumStrats(m_nfgPanel->GetRowPlayer());
  return wxString(m_doc->GetNfgSupport().Strategies(m_nfgPanel->GetRowPlayer())[p_row+1]->Name().c_str(), *wxConvCurrent);
}

wxString gbtMatrixSheet::GetColLabelValue(int p_col)
{
  const gbtNfgSupport &support = m_doc->GetNfgSupport();
  int colPlayer = m_nfgPanel->GetColPlayer();
  int numStrats = support.NumStrats(colPlayer);
  int numPlayers = m_doc->GetNfg()->NumPlayers();
  int strategy = (p_col / numPlayers) + 1;

  return wxString(support.Strategies(colPlayer)[strategy]->Name().c_str(),
		  *wxConvCurrent);
}

wxString gbtMatrixSheet::GetCellValue(const wxSheetCoords &p_coords)
{
  int row = p_coords.GetRow();
  int col = p_coords.GetCol();

  if (row < 0 && col >= 0)  return GetColLabelValue(col);
  else if (row >= 0 && col < 0)  return GetRowLabelValue(row);
  else if (row < 0 && col < 0)  return wxT("");

  int rowPlayer = m_nfgPanel->GetRowPlayer();
  int colPlayer = m_nfgPanel->GetColPlayer();
  int numPlayers = m_doc->GetNfg()->NumPlayers();
  const gbtNfgSupport &support = m_doc->GetNfgSupport();
  int numRowStrats = support.NumStrats(rowPlayer);
  int numColStrats = support.NumStrats(colPlayer);

  gArray<int> strategy(m_nfgPanel->GetContingency());
  strategy[rowPlayer] = row + 1;
  strategy[colPlayer] = (col / numPlayers) + 1;
    
  StrategyProfile profile(*m_doc->GetNfg());
  for (int pl = 1; pl <= strategy.Length(); pl++) {
    profile.Set(pl, support.GetStrategy(pl, strategy[pl]));
  }

  gbtNfgOutcome *outcome = m_doc->GetNfg()->GetOutcome(profile);
  int player = (col % numPlayers) + 1;
  return wxString(m_doc->GetNfg()->GetPayoffText(outcome, player).c_str(),
		  *wxConvCurrent);
}

void gbtMatrixSheet::SetCellValue(const wxSheetCoords &p_coords,
				  const wxString &p_value)
{
  if (IsCornerLabelCell(p_coords))  {
    return;
  }
  else if (IsRowLabelCell(p_coords)) {
    int rowStrat = p_coords.GetRow();
    gbtNfgGame *game = m_doc->GetNfg();
    gbtNfgPlayer *player = game->GetPlayer(m_nfgPanel->GetRowPlayer());
    player->GetStrategy(rowStrat + 1)->SetName((const char *) p_value.mb_str());
    m_doc->UpdateViews(true);
    return;
  }
  else if (IsColLabelCell(p_coords)) {
    gbtNfgGame *game = m_doc->GetNfg();
    int colStrat = p_coords.GetCol() / game->NumPlayers();
    gbtNfgPlayer *player = game->GetPlayer(m_nfgPanel->GetColPlayer());
    player->GetStrategy(colStrat + 1)->SetName((const char *) p_value.mb_str());
    m_doc->UpdateViews(true);
    return;
  }

  int rowPlayer = m_nfgPanel->GetRowPlayer();
  int colPlayer = m_nfgPanel->GetColPlayer();
  int rowStrat = p_coords.GetRow();
  int colStrat = p_coords.GetCol() / m_doc->GetNfg()->NumPlayers();
  int player = p_coords.GetCol() % m_doc->GetNfg()->NumPlayers() + 1;

  gbtNfgGame *game = m_doc->GetNfg();
  const gbtNfgSupport &support = m_doc->GetNfgSupport();

  gArray<int> strategy(m_nfgPanel->GetContingency());
  strategy[rowPlayer] = p_coords.GetRow() + 1;
  strategy[colPlayer] = (p_coords.GetCol() / game->NumPlayers()) + 1;
    
  StrategyProfile profile(*game);
  for (int pl = 1; pl <= game->NumPlayers(); pl++) {
    profile.Set(pl, support.GetStrategy(pl, strategy[pl]));
  }

  gbtNfgOutcome *outcome = game->GetOutcome(profile);
  if (!outcome) {
    outcome = game->NewOutcome();
    game->SetOutcome(profile, outcome);
  }

  game->SetPayoff(outcome, player, (const char *) p_value.mb_str());
  m_doc->UpdateViews(true);
}

wxSheetCellAttr gbtMatrixSheet::GetAttr(const wxSheetCoords &p_coords,
					wxSheetAttr_Type) const 
{
  if (IsRowLabelCell(p_coords)) {
    wxSheetCellAttr attr(GetSheetRefData()->m_defaultRowLabelAttr);
    attr.SetFont(wxFont(12, wxSWISS, wxNORMAL, wxBOLD));
    attr.SetAlignment(wxALIGN_CENTER, wxALIGN_CENTER);
    attr.SetOrientation(wxHORIZONTAL);
    attr.SetReadOnly(FALSE);
    attr.SetForegroundColour(m_doc->GetStyle().GetPlayerColor(m_nfgPanel->GetRowPlayer()));
    return attr;
  }
  else if (IsColLabelCell(p_coords)) {
    wxSheetCellAttr attr(GetSheetRefData()->m_defaultColLabelAttr);
    attr.SetFont(wxFont(12, wxSWISS, wxNORMAL, wxBOLD));
    attr.SetAlignment(wxALIGN_CENTER, wxALIGN_CENTER);
    attr.SetOrientation(wxHORIZONTAL);
    attr.SetReadOnly(FALSE);
    attr.SetForegroundColour(m_doc->GetStyle().GetPlayerColor(m_nfgPanel->GetColPlayer()));
    return attr;
  }
  else if (IsCornerLabelCell(p_coords)) {
    return GetSheetRefData()->m_defaultCornerLabelAttr;
  }

  wxSheetCellAttr attr(GetSheetRefData()->m_defaultGridCellAttr);
  attr.SetFont(wxFont(10, wxSWISS, wxNORMAL, wxBOLD));
  attr.SetAlignment(wxALIGN_CENTER, wxALIGN_CENTER);
  attr.SetOrientation(wxHORIZONTAL);
  attr.SetRenderer(wxSheetCellRenderer(new gbtRationalRendererRefData()));
  switch (p_coords.GetCol() % m_doc->GetNfg()->NumPlayers()) {
  case 0:
    attr.SetForegroundColour(m_doc->GetStyle().GetPlayerColor(m_nfgPanel->GetRowPlayer()));
    break;
  case 1:
    attr.SetForegroundColour(m_doc->GetStyle().GetPlayerColor(m_nfgPanel->GetColPlayer()));
    break;
  default:
    attr.SetForegroundColour(m_doc->GetStyle().GetPlayerColor(m_nfgPanel->GetTablePlayer(p_coords.GetCol() % m_doc->NumPlayers() - 1)));
    break;
  }
			   
  attr.SetReadOnly(false);
  return attr;
}

void gbtMatrixSheet::OnUpdate(void)
{
  if (!m_doc->GetNfg())  return;

  BeginBatch();
  int stratRows = GetNumberRows();
  int stratCols = GetNumberCols();

  gbtNfgGame *nfg = m_doc->GetNfg();
  const gbtNfgSupport &support = m_doc->GetNfgSupport();

  int dim = support.NumStrats(m_nfgPanel->GetRowPlayer());
	     
  if (dim < stratRows) {
    DeleteRows(0, stratRows - dim);
  }
  else if (dim > stratRows) {
    InsertRows(0, dim - stratRows); 
  }

  dim = (support.NumStrats(m_nfgPanel->GetColPlayer()) * nfg->NumPlayers());
  if (dim < stratCols) {
    DeleteCols(0, stratCols - dim);
  }
  else if (dim > stratCols) {
    InsertCols(0, dim - stratCols);
  }

  EndBatch();
  AdjustScrollbars();
}

//=========================================================================
// Below, we override some wxSheet internals in such a way that the
// column labels get drawn as if "merged".
//=========================================================================

static bool wxRectIsEmpty(const wxRect &rect)
{
  return (rect.width < 1) || (rect.height < 1);
}

void gbtMatrixSheet::DrawCellBorder( wxDC& dc, const wxSheetCoords& coords )
{
  wxRect rect(CellToRect( coords ));
  if ( wxRectIsEmpty(rect) )  // !IsCellShown
    return;

  dc.SetPen( wxPen(GetGridLineColour(), 1, wxSOLID) );
  // right hand border
  // the right border gets drawn only for the last column
  if ((coords.GetCol() + 1) % m_doc->GetNfg()->NumPlayers() == 0) {
    dc.DrawLine( rect.x + rect.width, rect.y,
		 rect.x + rect.width, rect.y + rect.height + 1 );
  }

  // bottom border
  dc.DrawLine( rect.x - 1,          rect.y + rect.height,
	       rect.x + rect.width, rect.y + rect.height);
}

//
// This overrides cell rectangle calculations for column labels
//
wxRect gbtMatrixSheet::CellToRect(const wxSheetCoords &p_coords,
				  bool p_getDeviceRect) const
{
  if (IsColLabelCell(p_coords)) {
    int virtcol = p_coords.GetCol() - (p_coords.GetCol() % m_doc->GetNfg()->NumPlayers());
    wxSheetBlock block(-1, virtcol, 1, m_doc->GetNfg()->NumPlayers());

    wxRect rect(GetColLeft(block.GetLeft()), GetRowTop(block.GetTop()), 0, 0);
    rect.width  = GetColRight( block.GetRight() ) - rect.x; // + 1;
    rect.height = GetRowBottom(block.GetBottom()) - rect.y; // + 1;
    
    if (GridLinesEnabled()) {
      rect.width--;
      rect.height--;
    }
    
    int dummy;
    CalcScrolledPosition(0, rect.y, &dummy, &rect.y);
    return rect;

  }
  else {
    return wxSheet::CellToRect(p_coords, p_getDeviceRect);
  }
}

// This is identical to the wxSheet implementation, except the
// right sides of only odd-numbered columns are drawn
void gbtMatrixSheet::DrawColLabels( wxDC& dc, const wxArrayInt& cols )
{
  size_t i, numLabels = cols.GetCount();
  if (GetNumberCols() == 0 || !numLabels ) return;

  for ( i = 0; i < numLabels; i++ )
    DrawCell( dc, wxSheetCoords(-1, cols[i]) );

  // Now draw the dividing lines
  dc.SetPen(wxPen(GetLabelOutlineColour(), 1, wxSOLID));
  int left, dummy;
  CalcUnscrolledPosition(0, 0, &left, &dummy);
  int right = GetColRight(cols[numLabels-1]); 
  // top of col labels
  dc.DrawLine(left, 0, right, 0);
  // bottom of col labels
  dc.DrawLine(left, GetColLabelHeight()-1, right, GetColLabelHeight()-1);
  // draw rights
  for (i = 0; i < numLabels; i++) {
    if ((cols[i] + 1) % m_doc->GetNfg()->NumPlayers() == 0) {
      right = GetColRight(cols[i])-1;
      dc.DrawLine(right, 0, right, GetColLabelHeight()-1);
    }
  }
}

//======================================================================
//                   class gbtNfgPanel: Member functions
//======================================================================

BEGIN_EVENT_TABLE(gbtNfgPanel, wxPanel)
  EVT_ROW_PLAYER_CHANGE(gbtNfgPanel::OnSetRowPlayer)
  EVT_COL_PLAYER_CHANGE(gbtNfgPanel::OnSetColPlayer)
  EVT_MENU(GBT_MENU_TOOLS_DOMINANCE, gbtNfgPanel::OnToolsDominance)
END_EVENT_TABLE()

gbtNfgPanel::gbtNfgPanel(wxWindow *p_parent, gbtGameDocument *p_doc)
  : wxPanel(p_parent, -1), gbtGameView(p_doc)
{
  m_dominanceToolbar = new gbtStrategyDominanceToolbar(this, m_doc);

  m_rowPlayerCtrl = new gbtPlayerLabelCtrl(this, m_doc, true);
  m_rowPlayerCtrl->SetPlayer(1);
  m_colPlayerCtrl = new gbtPlayerLabelCtrl(this, m_doc, false);
  m_colPlayerCtrl->SetPlayer(2);

  m_grid = new gbtMatrixSheet(this, -1, m_doc);

  m_playerToolbar = new gbtTablePlayerToolbar(this, m_doc);

  wxPanel *fooPanel = new wxPanel(this, -1);
  wxFlexGridSizer *gridSizer = new wxFlexGridSizer(2, 2);
  gridSizer->AddGrowableRow(1);
  gridSizer->AddGrowableCol(1);
  gridSizer->Add(fooPanel, 0, wxALL, 0);
  gridSizer->Add(m_colPlayerCtrl, 1, wxEXPAND, 0);
  gridSizer->Add(m_rowPlayerCtrl, 1, wxEXPAND, 0);
  gridSizer->Add(m_grid, 1, wxEXPAND, 0);

  wxBoxSizer *playerSizer = new wxBoxSizer(wxHORIZONTAL);
  playerSizer->Add(m_playerToolbar, 0, wxEXPAND, 0);
  playerSizer->Add(gridSizer, 1, wxEXPAND, 0);

  wxBoxSizer *topSizer = new wxBoxSizer(wxVERTICAL);
  topSizer->Add(m_dominanceToolbar, 0, wxEXPAND, 0);
  topSizer->Show(m_dominanceToolbar, false);
  topSizer->Add(playerSizer, 1, wxEXPAND, 0);
  SetSizer(topSizer);
  Layout();
}

void gbtNfgPanel::SetStrategy(int player, int strategy)
{
  m_contingency[player] = strategy;
  dynamic_cast<gbtMatrixSheet *>(m_grid)->OnUpdate();
}

void gbtNfgPanel::OnSetRowPlayer(wxCommandEvent &p_event)
{
  if (p_event.GetInt() == m_colPlayerCtrl->GetPlayer()) {
    m_colPlayerCtrl->SetPlayer(m_rowPlayerCtrl->GetPlayer());
  }
  m_rowPlayerCtrl->SetPlayer(p_event.GetInt());
  OnUpdate();
}

void gbtNfgPanel::OnSetColPlayer(wxCommandEvent &p_event)
{
  if (p_event.GetInt() == m_rowPlayerCtrl->GetPlayer()) {
    m_rowPlayerCtrl->SetPlayer(m_colPlayerCtrl->GetPlayer());
  }
  m_colPlayerCtrl->SetPlayer(p_event.GetInt());
  OnUpdate();
}

void gbtNfgPanel::OnToolsDominance(wxCommandEvent &p_event)
{
  GetSizer()->Show(m_dominanceToolbar, p_event.IsChecked(), true);
  GetSizer()->Layout();
}

void gbtNfgPanel::OnUpdate(void)
{ 
  if (!m_doc->GetNfg())  return;
 
  if (m_contingency.Length() != m_doc->GetNfg()->NumPlayers()) {
    m_contingency = gArray<int>(m_doc->GetNfg()->NumPlayers());
    for (int pl = 1; pl <= m_contingency.Length(); m_contingency[pl++] = 1);
  }

  m_grid->EnableEditing(!m_doc->GetEfg());

  dynamic_cast<gbtMatrixSheet *>(m_grid)->OnUpdate();
  m_grid->AutoSizeRows();
  m_grid->AutoSizeColumns();
  m_grid->AdjustScrollbars();
  m_grid->ForceRefresh();

  GetSizer()->Layout();
}

int gbtNfgPanel::GetTablePlayer(int index) const
{
  int pl = 0;
  while (index > 0) {
    pl++;
    if (pl != GetRowPlayer() && pl != GetColPlayer()) {
      index--;
    }
  }
  return pl;
}


DEFINE_EVENT_TYPE(GBT_ROW_PLAYER_CHANGE)
DEFINE_EVENT_TYPE(GBT_COL_PLAYER_CHANGE)

