//
// $Source: /cvsroot/gambit/gambit/sources/tools/enummixed/Attic/enummixed.cc,v $
// $Date: 2005/06/14 17:26:02 $
// $Revision: 1.1.2.3 $
//
// DESCRIPTION:
// Compute Nash equilibria via Mangasarian's algorithm
//
// This file is part of Gambit
// Copyright (c) 2002, The Gambit Project
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
//

#include <iostream>

#include "vertenum.imp"

#include "game/nfg.h"
#include "game/nfgiter.h"

template <class T> bool EqZero(const T &x)
{
  T eps;
  gEpsilon(eps, 12);
  return (x <= eps && x >= -eps);
}     

int m_stopAfter = 0;

template <class T>
void PrintProfile(std::ostream &p_stream,
		  const MixedProfile<T> &p_profile)
{
  for (int i = 1; i <= p_profile.Length(); i++) {
    p_stream << p_profile[i];
    if (i < p_profile.Length()) {
      p_stream << ',';
    }
  }

  p_stream << std::endl;
}

template <class T> void Solve(const gbtNfgGame &p_nfg, const T &)
{
  gList<gVector<T> > key1, key2;  
  gList<int> node1, node2;   // IDs of each component of the extreme equilibria

  NfgIter iter(p_nfg);

  gRational min = MinPayoff(p_nfg);
  if (min > gRational(0)) {
    min = gRational(0);
  }
  min -= gRational(1);

  gRational max = MaxPayoff(p_nfg);
  if (max < gRational(0)) {
    max = gRational(0);
  }

  gRational fac(1, max - min);

  // Construct matrices A1, A2
  gMatrix<T> A1(1, p_nfg.NumStrats(1), 1, p_nfg.NumStrats(2));
  gMatrix<T> A2(1, p_nfg.NumStrats(2), 1, p_nfg.NumStrats(1));

  for (int i = 1; i <= p_nfg.NumStrats(1); i++) {
    for (int j = 1; j <= p_nfg.NumStrats(2); j++) {
      A1(i, j) = fac * (p_nfg.Payoff(iter.GetOutcome(), 1) - min);
      A2(j, i) = fac * (p_nfg.Payoff(iter.GetOutcome(), 2) - min);
      iter.Next(2);
    }
    iter.Next(1);
  }

  // Construct vectors b1, b2
  gVector<T> b1(1, p_nfg.NumStrats(1)), b2(1, p_nfg.NumStrats(2));
  b1 = (T) -1;
  b2 = (T) -1;

  // enumerate vertices of A1 x + b1 <= 0 and A2 x + b2 <= 0
  VertEnum<T> poly1(A1, b1);
  VertEnum<T> poly2(A2, b2);

  const gList<BFS<T> > &verts1(poly1.VertexList());
  const gList<BFS<T> > &verts2(poly2.VertexList());
  int v1 = verts1.Length();
  int v2 = verts2.Length();

  gArray<int> vert1id(v1);
  gArray<int> vert2id(v2);
  for (int i = 1; i <= vert1id.Length(); vert1id[i++] = 0);
  for (int i = 1; i <= vert2id.Length(); vert2id[i++] = 0);

  int i = 0;
  int id1 = 0, id2 = 0;

  try {
    for (int i2 = 2; i2 <= v2; i2++) {
      BFS<T> bfs1 = verts2[i2];
      i++;
      for (int i1 = 2; i1 <= v1; i1++) {
	BFS<T> bfs2 = verts1[i1];
	
	// check if solution is nash 
	// need only check complementarity, since it is feasible
	bool nash = true;
	for (int k = 1; nash && k <= p_nfg.NumStrats(1); k++) {
	  if (bfs1.IsDefined(k) && bfs2.IsDefined(-k)) {
	    nash = nash && EqZero(bfs1(k) * bfs2(-k));
	  }
	}

	for (int k = 1; nash && k <= p_nfg.NumStrats(2); k++) {
	  if (bfs2.IsDefined(k) && bfs1.IsDefined(-k)) {
	    nash = nash && EqZero(bfs2(k) * bfs1(-k));
	  }
	}

	if (nash) {
	  MixedProfile<T> profile(p_nfg);
	  T sum = (T) 0;
	  for (int k = 1; k <= p_nfg.NumStrats(1); k++) {
	    profile(1, k) = (T) 0;
	    if (bfs1.IsDefined(k)) {
	      profile(1,k) = -bfs1(k);
	      sum += profile(1,k);
	    }
	  } 
	  
	  for (int k = 1; k <= p_nfg.NumStrats(1); k++) {
	    if (bfs1.IsDefined(k)) { 
	      profile(1,k) /= sum;
	    }
	  }
	  
	  sum = (T) 0;
	  for (int k = 1; k <= p_nfg.NumStrats(2); k++) {
	    profile(2,k) = (T) 0;
	    if (bfs2.IsDefined(k)) {
	      profile(2,k) =-bfs2(k);
	      sum += profile(2,k);
	    }
	  } 
	  
	  for (int k = 1; k <= p_nfg.NumStrats(2); k++) {
	    if (bfs2.IsDefined(k)) { 
	      profile(2,k) /= sum;
	    }
	  }

	  PrintProfile(std::cout, profile);
	  
	  // note: The keys give the mixed strategy associated with each node. 
	  //       The keys should also keep track of the basis
	  //       As things stand now, two different bases could lead to
	  //       the same key... BAD!
	  if (vert1id[i1] == 0) {
	    id1++;
	    vert1id[i1] = id1;
	    key2.Append(profile.GetRow(2));
	  }
	  if (vert2id[i2] == 0) {
	    id2++;
	    vert2id[i2] = id2;
	    key1.Append(profile.GetRow(1));
	  }
	  node1.Append(vert2id[i2]);
	  node2.Append(vert1id[i1]);
	}
      }
    }
  }
  catch (...) {
    // stop processing, just return equilibria found (if any)
  }
}

#ifdef UNUSED
template <class T> void EnumModule<T>::GetCliques(gOutput &p_stream) const
{
  int n = node1.Length();
  assert(node2.Length() == n);

  gArray<edge> edgelist(n);

  p_stream << "\nKey:\nPlayer 1:";
  for (int i = 1; i <= key1.Length(); i++) {
    p_stream << "\n" << i << ": " << key1[i];
  }
  p_stream << "\nPlayer 2:";
  for (int i = 1; i <= key2.Length(); i++) {
    p_stream << "\n" << i << ": " << key2[i];
  }
  p_stream << "\nExtreme equilibria:";
  for (int i = 1; i <= n; i++) {
    edgelist[i].node1 = node1[i];
    edgelist[i].node2 = node2[i];
    p_stream << "\n" << node1[i] << " " << node2[i];
  }
  EnumCliques clique(edgelist, v2+1, v1+1);
}

#endif  // UNUSED

int main(int, char *argv[])
{
  gbtNfgGame *nfg;

  try {
    nfg = ReadNfg(std::cin);
  }
  catch (...) {
    return 1;
  }

  if (nfg->NumPlayers() != 2) {
    return 1;
  }

  try {
    Solve(*nfg, 0.0);
    return 0;
  }
  catch (...) {
    return 1;
  }
}










