//
// $Source: /cvsroot/gambit/gambit/sources/tools/lp/Attic/nfglp.cc,v $
// $Date: 2005/06/13 02:32:36 $
// $Revision: 1.1.2.2 $
//
// DESCRIPTION:
// Implementation of algorithm to compute mixed strategy equilibria
// of constant sum normal form games via linear programming
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
//

#include <iostream>

#include "game/nfg.h"
#include "game/mixed.h"
#include "game/nfgiter.h"
#include "lpsolve.h"

template <class T> class nfgLp {
private:
  int Add_BFS(const gbtNfgSupport &, /*const*/ LPSolve<T> &B,
	      gList<BFS<T> > &);
  void GetSolutions(const gbtNfgSupport &, const gList<BFS<T> > &,
		    const T &) const;

public:
  nfgLp(void);
  virtual ~nfgLp() { }

  void Solve(const gbtNfgSupport &);
};

//-------------------------------------------------------------------------
//                      nfgLp<T>: Member functions
//-------------------------------------------------------------------------

template <class T> nfgLp<T>::nfgLp(void)
{ }

template <class T> 
void nfgLp<T>::Solve(const gbtNfgSupport &p_support)
{
  BFS<T> cbfs((T) 0);
  
  if (p_support.Game()->NumPlayers() != 2 || !IsConstSum(*p_support.Game())) {
    return; 
  }
  
  gList<BFS<T> > bfsList;
  
  int i,j,m,k;
  
  m = p_support.NumStrats(1);
  k = p_support.NumStrats(2);
  gMatrix<T> A(1,k+1,1,m+1);
  gVector<T> b(1,k+1);
  gVector<T> c(1,m+1);
  NfgIter iter(p_support);

  T minpay = MinPayoff(*p_support.Game()) - gRational(1);

  for (i = 1; i <= k; i++)  {
    for (j = 1; j <= m; j++)  {
      A(i, j) = gRational(minpay) - p_support.Game()->Payoff(iter.GetOutcome(), 1);
      iter.Next(1);
    }
    A(i,m+1) = (T)1;
    iter.Next(2);
  }
  for (j = 1;j<=m;j++)  {
    A(k+1,j)= (T)1;
  }
  A(k+1,m+1) = (T)0;

  b = (T)0;
  b[k+1] = (T)1;
  c = (T)0;
  c[m+1] = (T)1;

  LPSolve<T> LP(A,b,c,1);

  if (!LP.IsAborted()) {
    Add_BFS(p_support, LP, bfsList); 
  }

  GetSolutions(p_support, bfsList, LP.Epsilon());
}

template <class T> int nfgLp<T>::Add_BFS(const gbtNfgSupport &p_support,
					 /*const*/ LPSolve<T> &lp,
					 gList<BFS<T> > &p_list)
{
  BFS<T> cbfs((T) 0);

  // LPSolve<T>::GetAll() does not currently work correctly; for now,
  // LpSolve is restricted to returning only one equilibrium
  lp.OptBFS(cbfs);
  cbfs.Remove(p_support.Game()->NumStrats(1)+1);
  cbfs.Remove(-p_support.Game()->NumStrats(2)-1);
  if (p_list.Contains(cbfs))  return 0;
  p_list.Append(cbfs);
  return 1;
}

template <class T>
void PrintProfile(std::ostream &p_stream,
		  const MixedProfile<T> &p_profile)
{
  for (int i = 1; i <= p_profile.Length(); i++) {
    p_stream << p_profile[i];
    if (i < p_profile.Length()) {
      p_stream << ',';
    }
  }

  p_stream << std::endl;
}

template <class T>
void nfgLp<T>::GetSolutions(const gbtNfgSupport &p_support,
			    const gList<BFS<T> > &p_list,
			    const T &p_epsilon) const
{
  int index;
  int n1 = p_support.NumStrats(1);
  int n2=p_support.NumStrats(2);

  for (int i = 1; i <= p_list.Length(); i++)    {
    MixedProfile<T> profile(p_support);
    int j;
    for (j = 1; j <= n1; j++) 
      if (p_list[i].IsDefined(j))   
	profile(1, j) = p_list[i](j);
      else  profile(1, j) = (T) 0;

    for (j = 1; j <= n2; j++)
      if (p_list[i].IsDefined(-j))
	profile(2, j) = p_list[i](-j);
      else
	profile(2, j) = (T) 0;

    PrintProfile(std::cout, profile);
  }
}

int main(int, char *argv[])
{
  gbtNfgGame *nfg;

  try {
    nfg = ReadNfg(std::cin);
  }
  catch (...) {
    return 1;
  }

  nfgLp<double> algorithm;
  algorithm.Solve(*nfg);

  return 0;

}

